import pytest
from prodsys.models.production_system_data import ProductionSystemData
import prodsys.express as psx
from prodsys import runner


@pytest.fixture
def simulation_adapter() -> ProductionSystemData:
    t1 = psx.FunctionTimeModel("exponential", 0.8, 0, "t1")

    p1 = psx.ProductionProcess(t1, "p1", failure_rate=0.05)
    p2 = psx.ProductionProcess(t1, "p2", failure_rate=0.1)
    p3 = psx.ProductionProcess(t1, "p3")

    t3 = psx.FunctionTimeModel("constant", 0.1, ID="t3")

    tp = psx.TransportProcess(t3, "tp")

    rework_time_model = psx.FunctionTimeModel("constant", 1, ID="rework_time_model")
    rework_time_model2 = psx.FunctionTimeModel("constant", 5, ID="rework_time_model")
    rework_time_model3 = psx.FunctionTimeModel("constant", 2, ID="rework_time_model")

    rework_process = psx.ReworkProcess(rework_time_model, [p1], True, "rework_process")
    rework_process2 = psx.ReworkProcess(
        rework_time_model2, [p2], False, "rework_process2"
    )
    rework_process3 = psx.ReworkProcess(
        rework_time_model3, [p1], False, "rework_process3"
    )

    machine = psx.Resource([p1], [5, 0], 1, ID="machine")
    machine2 = psx.Resource([p2], [10, 0], 1, ID="machine2")
    machine3 = psx.Resource([p3], [5, 5], 1, ID="machine3")

    reworker = psx.Resource(
        [rework_process, rework_process2, rework_process3], [8, 0], 1, ID="reworker"
    )

    transport = psx.Resource([tp], [0, 0], 1, ID="transport")

    product1 = psx.Product([p1, p2, p3], tp, "product1")

    sink1 = psx.Sink(product1, [15, 0], "sink1")

    arrival_model_1 = psx.FunctionTimeModel("exponential", 1, ID="arrival_model_1")

    source1 = psx.Source(product1, arrival_model_1, [0, 0], ID="source_2")

    system = psx.ProductionSystem(
        [machine, machine2, machine3, reworker, transport], [source1], [sink1]
    )
    adapter = system.to_model()
    return adapter


def test_initialize_simulation(simulation_adapter: ProductionSystemData):
    runner_instance = runner.Runner(production_system_data=simulation_adapter)
    runner_instance.initialize_simulation()


def test_hashing(simulation_adapter: ProductionSystemData):
    hash_str = simulation_adapter.hash()
    assert hash_str == "dd57c021062f2a9ebc0db054086ca001"


def test_run_simulation(simulation_adapter: ProductionSystemData):
    runner_instance = runner.Runner(production_system_data=simulation_adapter)
    runner_instance.initialize_simulation()
    runner_instance.run(2000)
    runner_instance.print_results()
    assert runner_instance.env.now == 2000
    post_processor = runner_instance.get_post_processor()
    for kpi in post_processor.throughput_and_output_KPIs:
        if kpi.name == "output":
            assert kpi.product_type == "product1"
            assert kpi.value > 1940 and kpi.value < 2000
    for kpi in post_processor.machine_state_KPIS:
        if kpi.name == "productive_time" and kpi.resource == "machine":
            assert kpi.value < 82 and kpi.value > 76

        if kpi.name == "productive_time" and kpi.resource == "transport":
            assert kpi.value > 73 and kpi.value < 82

        if kpi.name == "productive_time" and kpi.resource == "reworker":
            assert kpi.value > 14 and kpi.value < 16

    for kpi in post_processor.WIP_KPIs:
        if kpi.name == "WIP" and kpi.product_type == "product1":
            assert kpi.value < 16 and kpi.value > 13

    for kpi in post_processor.aggregated_throughput_time_KPIs:
        if kpi.name == "throughput_time":
            assert kpi.value < 16 and kpi.value > 13



from prodsys.simulation import request
from prodsys.models.dependency_data import DependencyType, LotDependencyData
from prodsys.simulation.dependency import Dependency
from prodsys.models.resource_data import ResourceType
from prodsys.simulation.entities.lot import Lot

class LotHandler:

    def _get_lot_dependency(self, process_request: request.Request) -> Dependency:
        for dependency in process_request.required_dependencies:
            if dependency.data.dependency_type == DependencyType.LOT:
                return dependency
        return None

    def _get_lot_dependency_data(self, process_request: request.Request) -> LotDependencyData:
        lot_dependencies = []
        for dependency in process_request.required_dependencies:
            if dependency.data.dependency_type == DependencyType.LOT:
                lot_dependencies.append(dependency.data)
        if len(lot_dependencies) == 0:
            return None
        if len(lot_dependencies) == 1:
            return lot_dependencies[0]
        # if there are multiple lot dependencies, we need to combine them into one lot dependency
        combined_lot_dependency = LotDependencyData(
            ID="combined_lot_dependency",
            description="Combined lot dependency",
            dependency_type=DependencyType.LOT,
            min_lot_size=min(lot_dependency.min_lot_size for lot_dependency in lot_dependencies),
            max_lot_size=max(lot_dependency.max_lot_size for lot_dependency in lot_dependencies),
        )
        return combined_lot_dependency

    def lot_required(self, process_request: request.Request) -> bool:
        if process_request.request_type not in [request.RequestType.PRODUCTION, request.RequestType.TRANSPORT, request.RequestType.PROCESS_MODEL]:
            return False

        if process_request.request_type == request.RequestType.PROCESS_MODEL and process_request.resource.data.resource_type == ResourceType.SYSTEM:
            return False
        
        if not self._get_lot_dependency_data(process_request):
            return False
        return True

    def _request_matches(self, process_request: request.Request, potential_lot_request: request.Request) -> bool:
        if process_request.request_type == request.RequestType.PRODUCTION or process_request.request_type == request.RequestType.PROCESS_MODEL:
            return process_request.process == potential_lot_request.process
        elif process_request.request_type == request.RequestType.TRANSPORT:
            return process_request.process == potential_lot_request.process and process_request.origin_queue == potential_lot_request.origin_queue and process_request.target_queue == potential_lot_request.target_queue
        else:
            return False

    def _get_possible_requests_for_lot(self, process_request: request.Request) -> list[request.Request]:
        possible_requests_for_lot = []
        for open_request in process_request.resource.controller.requests:
            if open_request is process_request:
                continue
            if self._request_matches(process_request, open_request):
                possible_requests_for_lot.append(open_request)
        return possible_requests_for_lot
    
    def is_lot_feasible(self, process_request: request.Request) -> bool:
        lot_dependency = self._get_lot_dependency_data(process_request)
        if lot_dependency is None:
            return True
        if process_request.resource.data.capacity < lot_dependency.min_lot_size:
            raise ValueError(f"The capacity of the resource {process_request.resource.data.ID} is smaller than the min lot size {lot_dependency.min_lot_size}")
        if process_request.resource.get_free_capacity() < lot_dependency.min_lot_size:
            return False
        # if process_request.target_queue.is_full or process_request.target_queue.free_space() < lot_dependency.min_lot_size: # avoids that reservations don't go through
        #     return False
        if process_request.request_type == request.RequestType.TRANSPORT:
            if process_request.target_queue.is_full or process_request.target_queue.free_space() < lot_dependency.min_lot_size:
                return False
        possible_requests_for_lot = self._get_possible_requests_for_lot(process_request)
        num_possible = len(possible_requests_for_lot)
        return num_possible >= lot_dependency.min_lot_size - 1


    def _get_requests_to_fill_lot(self, process_request: request.Request, lot_dependency: LotDependencyData, possible_requests_for_lot: list[request.Request]) -> list[request.Request]:
        if process_request.resource.get_free_capacity() < lot_dependency.max_lot_size:
            max_requests_to_fill_lot = process_request.resource.get_free_capacity() - 1
        elif process_request.request_type == request.RequestType.TRANSPORT and process_request.target_queue.free_space() < lot_dependency.max_lot_size:
            max_requests_to_fill_lot = process_request.target_queue.free_space() - 1
        else:
            max_requests_to_fill_lot = lot_dependency.max_lot_size - 1
        num_requests_to_fill_lot = 0
        if len(possible_requests_for_lot) < max_requests_to_fill_lot:
            num_requests_to_fill_lot = len(possible_requests_for_lot)
        else:
            num_requests_to_fill_lot = max_requests_to_fill_lot
        if num_requests_to_fill_lot < 0:
            raise ValueError(f"The number of requests to fill the lot is negative: {num_requests_to_fill_lot}")
        return possible_requests_for_lot[:num_requests_to_fill_lot]

    def get_lot_request(self, process_request: request.Request) -> request.Request:
        lot_dependency = self._get_lot_dependency_data(process_request)
        if lot_dependency is None:
            return [process_request]
        possible_requests_for_lot = self._get_possible_requests_for_lot(process_request)
        # use control policy to sort the requests
        process_request.resource.controller.control_policy(possible_requests_for_lot)
        requests_to_fill_lot = self._get_requests_to_fill_lot(process_request, lot_dependency, possible_requests_for_lot)
        for lot_request in requests_to_fill_lot:
            process_request.resource.controller.requests.remove(lot_request)
        lot_requests = [process_request] + requests_to_fill_lot
        lot_entities = [request.entity for request in lot_requests]
        all_completed_events = [request.completed for request in lot_requests]

        lot = Lot(
            all_completed_events=all_completed_events,
            entities=lot_entities,
            resolved_dependency=self._get_lot_dependency(process_request),
            required_dependencies=process_request.required_dependencies,
        )
        process_request.entity = lot
        process_request.required_dependencies = lot.dependencies
        return process_request


#!/usr/bin/env python3
"""
Tests for ProcessModel and SystemResource features.

This module contains tests for the new ProcessModel and SystemResource functionality.
"""

import prodsys.express as psx


def test_process_models_creation():
    """Test that process models can be created and converted to data models."""
    # Test ProcessModel
    tm = psx.FunctionTimeModel("normal", 1.0, 0.1, "tm1")
    process_model = psx.ProcessModel(
        time_model=tm,
        process_ids=["p1", "p2"],
        adjacency_matrix={"p1": ["p2"], "p2": []},
        ID="pm1"
    )
    data_model = process_model.to_model()
    assert data_model.ID == "pm1"
    assert data_model.process_ids == ["p1", "p2"]
    assert data_model.adjacency_matrix == {"p1": ["p2"], "p2": []}

    # Test SequentialProcess
    sequential_process = psx.SequentialProcess(
        time_model=tm,
        process_ids=["p1", "p2", "p3"],
        ID="sp1"
    )
    seq_data_model = sequential_process.to_model()
    assert seq_data_model.ID == "sp1"
    assert seq_data_model.process_ids == ["p1", "p2", "p3"]
    expected_adjacency = {"p1": ["p2"], "p2": ["p3"], "p3": []}
    assert seq_data_model.adjacency_matrix == expected_adjacency

    # Test LoadingProcess
    loading_process = psx.LoadingProcess(
        time_model=tm,
        dependency_type="before",
        can_be_chained=True,
        ID="lp1"
    )
    lp_data_model = loading_process.to_model()
    assert lp_data_model.ID == "lp1"
    assert lp_data_model.dependency_type == "before"
    assert lp_data_model.can_be_chained is True


def test_system_resource_creation():
    """Test that system resources can be created and converted to data models."""
    tm = psx.FunctionTimeModel("normal", 1.0, 0.1, "tm1")
    p1 = psx.ProductionProcess(tm, "p1")
    
    system_resource = psx.SystemResource(
        processes=[p1],
        location=[10, 10],
        subresource_ids=["r1", "r2"],
        ID="sr1"
    )
    
    data_model = system_resource.to_model()
    assert data_model.ID == "sr1"
    assert data_model.subresource_ids == ["r1", "r2"]

def test_product_with_process_model():
    """Test that products can use process models."""
    tm = psx.FunctionTimeModel("normal", 1.0, 0.1, "tm1")
    tp_tm = psx.DistanceTimeModel(60, 0.1, "manhattan", "tp_tm")

    process_model = psx.ProcessModel(
        time_model=tm,
        process_ids=["p1", "p2"],
        adjacency_matrix={"p1": ["p2"], "p2": []},
        ID="pm1"
    )
    
    transport_process = psx.TransportProcess(tp_tm, "tp1")
    
    product = psx.Product(
        process=process_model,
        transport_process=transport_process,
        ID="product1"
    )
    
    data_model = product.to_model()
    assert data_model.ID == "product1"
    assert len(data_model.processes) == 2  # adjacency matrix has 2 processes: p1 and p2
    assert "p1" in data_model.processes
    assert "p2" in data_model.processes


def test_sequential_process_auto_adjacency():
    """Test that SequentialProcess automatically generates adjacency matrix."""
    tm = psx.FunctionTimeModel("normal", 1.0, 0.1, "tm1")
    
    # Test with 3 processes
    seq_process = psx.SequentialProcess(
        time_model=tm,
        process_ids=["p1", "p2", "p3"],
        ID="seq1"
    )
    
    expected_adjacency = {"p1": ["p2"], "p2": ["p3"], "p3": []}
    assert seq_process.adjacency_matrix == expected_adjacency
    
    # Test with 2 processes
    seq_process2 = psx.SequentialProcess(
        time_model=tm,
        process_ids=["p1", "p2"],
        ID="seq2"
    )
    
    expected_adjacency2 = {"p1": ["p2"], "p2": []}
    assert seq_process2.adjacency_matrix == expected_adjacency2


def test_loading_process_dependency_types():
    """Test that LoadingProcess supports different dependency types."""
    tm = psx.FunctionTimeModel("normal", 1.0, 0.1, "tm1")
    
    # Test "before" dependency
    lp_before = psx.LoadingProcess(
        time_model=tm,
        dependency_type="before",
        can_be_chained=True,
        ID="lp_before"
    )
    assert lp_before.dependency_type == "before"
    
    # Test "after" dependency
    lp_after = psx.LoadingProcess(
        time_model=tm,
        dependency_type="after",
        can_be_chained=False,
        ID="lp_after"
    )
    assert lp_after.dependency_type == "after"
    assert lp_after.can_be_chained is False
    
    # Test "parallel" dependency
    lp_parallel = psx.LoadingProcess(
        time_model=tm,
        dependency_type="parallel",
        can_be_chained=True,
        ID="lp_parallel"
    )
    assert lp_parallel.dependency_type == "parallel"


def test_process_model_can_contain_other_models():
    """Test that ProcessModel can be configured to contain other models."""
    tm = psx.FunctionTimeModel("normal", 1.0, 0.1, "tm1")
    
    # Test with can_contain_other_models=True
    pm_with_models = psx.ProcessModel(
        time_model=tm,
        process_ids=["p1", "p2"],
        adjacency_matrix={"p1": ["p2"], "p2": []},
        can_contain_other_models=True,
        ID="pm_with_models"
    )
    assert pm_with_models.can_contain_other_models is True
    
    # Test with can_contain_other_models=False (default)
    pm_without_models = psx.ProcessModel(
        time_model=tm,
        process_ids=["p1", "p2"],
        adjacency_matrix={"p1": ["p2"], "p2": []},
        ID="pm_without_models"
    )
    assert pm_without_models.can_contain_other_models is False


def test_system_resource_without_ports():
    """Test that SystemResource can be created without system ports."""
    tm = psx.FunctionTimeModel("normal", 1.0, 0.1, "tm1")
    p1 = psx.ProductionProcess(tm, "p1")
    
    system_resource = psx.SystemResource(
        processes=[p1],
        location=[10, 10],
        subresource_ids=["r1", "r2"],
        ID="sr1"
    )
    
    data_model = system_resource.to_model()
    assert data_model.ID == "sr1"
    assert data_model.subresource_ids == ["r1", "r2"]


def test_data_model_conversion():
    """Test that all new models can be converted to data models correctly."""
    tm = psx.FunctionTimeModel("normal", 1.0, 0.1, "tm1")
    
    # Test ProcessModel conversion
    pm = psx.ProcessModel(
        time_model=tm,
        process_ids=["p1", "p2"],
        adjacency_matrix={"p1": ["p2"], "p2": []},
        ID="pm1"
    )
    pm_data = pm.to_model()
    assert pm_data.type.value == "ProcessModels"
    
    # Test SequentialProcess conversion
    sp = psx.SequentialProcess(
        time_model=tm,
        process_ids=["p1", "p2"],
        ID="sp1"
    )
    sp_data = sp.to_model()
    assert sp_data.type.value == "ProcessModels"
    
    # Test LoadingProcess conversion
    lp = psx.LoadingProcess(
        time_model=tm,
        dependency_type="before",
        can_be_chained=True,
        ID="lp1"
    )
    lp_data = lp.to_model()
    assert lp_data.type.value == "LoadingProcesses"
    
    # Test SystemResource conversion
    sr = psx.SystemResource(
        processes=[pm],
        location=[10, 10],
        subresource_ids=["r1"],
        ID="sr1"
    )
    sr_data = sr.to_model()
    assert sr_data.ID == "sr1"
    assert sr_data.subresource_ids == ["r1"]


def test_system_resource_with_robot_simulation():
    """Test a complete simulation with SystemResource containing robot and machines."""
    
    # ========== TIME MODELS ==========
    tm_machine1 = psx.FunctionTimeModel(
        distribution_function="normal",
        location=1.8,
        scale=0.2,
        ID="tm_machine1"
    )
    tm_machine2 = psx.FunctionTimeModel(
        distribution_function="normal",
        location=2.2,
        scale=0.7,
        ID="tm_machine2"
    )
    tm_machine3 = psx.FunctionTimeModel(
        distribution_function="normal",
        location=2.0,
        scale=0.5,
        ID="tm_machine3"
    )
    
    tm_agv = psx.DistanceTimeModel(
        speed=60.0,
        reaction_time=0.1,
        metric="manhattan",
        ID="tm_agv"
    )
    tm_arrival = psx.FunctionTimeModel(
        distribution_function="exponential",
        location=2.5,
        ID="tm_arrival"
    )
    
    # ========== PROCESSES ==========
    machine1_process = psx.ProductionProcess(
        time_model=tm_machine1,
        ID="machine1_process"
    )
    machine2_process = psx.ProductionProcess(
        time_model=tm_machine2,
        ID="machine2_process"
    )
    machine3_process = psx.ProductionProcess(
        time_model=tm_machine3,
        ID="machine3_process"
    )
    
    agv_transport = psx.TransportProcess(
        time_model=tm_agv,
        ID="agv_transport"
    )
    
    # ========== PROCESS MODEL ==========
    cell_process_model = psx.ProcessModel(
        adjacency_matrix={
            "machine1_process": ["machine2_process"],
            "machine2_process": []
        },
        ID="cell_process_model"
    )
    
    product_process_model = psx.ProcessModel(
        can_contain_other_models=True,
        ID="product_process_model",
        adjacency_matrix={
            "machine3_process": ["cell_process_model"],
            "cell_process_model": []
        }
    )
    
    # ========== RESOURCES ==========
    robot = psx.Resource(
        processes=[agv_transport],
        location=[12, 10],
        capacity=1,
        ID="robot"
    )
    
    machine1 = psx.Resource(
        processes=[machine1_process],
        location=[12, 8],
        capacity=1,
        ID="machine1"
    )
    machine2 = psx.Resource(
        processes=[machine2_process],
        location=[12, 12],
        capacity=1,
        ID="machine2"
    )
    machine3 = psx.Resource(
        processes=[machine3_process],
        location=[12, 16],
        capacity=1,
        ID="machine3"
    )
    
    agv = psx.Resource(
        processes=[agv_transport],
        location=[0, 10],
        capacity=1,
        ID="agv"
    )
    
    # ========== SYSTEM RESOURCE (CELL) ==========
    manufacturing_cell = psx.SystemResource(
        processes=[cell_process_model],
        location=[10, 10],
        subresource_ids=["robot", "machine1", "machine2"],
        capacity=5,
        ID="manufacturing_cell"
    )
    
    # ========== PRODUCT ==========
    product = psx.Product(
        process=product_process_model,
        transport_process=agv_transport,
        ID="product"
    )
    
    # ========== SOURCES AND SINKS ==========
    source = psx.Source(product, tm_arrival, [0, 10], ID="source")
    sink = psx.Sink(product, [20, 10], ID="sink")
    
    # ========== PRODUCTION SYSTEM ==========
    system = psx.ProductionSystem(
        resources=[robot, machine1, machine2, machine3, manufacturing_cell, agv],
        sources=[source],
        sinks=[sink],
        ID="production_system"
    )
    
    # ========== VALIDATION AND SIMULATION ==========
    system.validate()
    system_data = system.to_model()
    
    # Validate system structure
    assert len(system_data.resource_data) == 6
    assert "manufacturing_cell" in [r.ID for r in system_data.resource_data]
    
    # Run simulation^
    system.run(time_range=1000)
    system.runner.print_results()
    
    # Get post processor for KPI validation
    post_processor = system.runner.get_post_processor()
    
    # Validate output and throughput KPIs (from terminal: Output=395, Throughput=0.397259)
    for kpi in post_processor.throughput_and_output_KPIs:
        if kpi.name == "output" and kpi.product_type == "product":
            assert kpi.value > 350 and kpi.value < 420, f"Output {kpi.value} out of expected range"
    # Validate WIP KPIs (from terminal: WIP=10.180611)
    for kpi in post_processor.WIP_KPIs:
        if kpi.name == "WIP" and kpi.product_type == "product":
            assert kpi.value > 7.5 and kpi.value < 12.5, f"WIP {kpi.value} out of expected range"
    
    # Validate throughput time (from terminal: 22.857539)
    for kpi in post_processor.aggregated_throughput_time_KPIs:
        if kpi.name == "throughput_time":
            assert kpi.value > 20.0 and kpi.value < 25.0, f"Throughput time {kpi.value} out of expected range"
    
    # Validate resource states (from terminal output)
    for kpi in post_processor.machine_state_KPIS:
        if kpi.name == "productive_time":
            if kpi.resource == "machine1":
                # Expected: ~71.7%
                assert kpi.value > 65 and kpi.value < 78, f"Machine1 productive time {kpi.value} out of expected range"
            elif kpi.resource == "machine2":
                # Expected: ~86.6%
                assert kpi.value > 80 and kpi.value < 92, f"Machine2 productive time {kpi.value} out of expected range"
            elif kpi.resource == "machine3":
                # Expected: ~80.3%
                assert kpi.value > 75 and kpi.value < 86, f"Machine3 productive time {kpi.value} out of expected range"
            elif kpi.resource == "agv":
                # Expected: ~32.3%
                assert kpi.value > 30 and kpi.value < 38, f"AGV productive time {kpi.value} out of expected range"
            elif kpi.resource == "robot":
                # Expected: ~64.0%
                assert kpi.value > 60 and kpi.value < 68, f"Robot productive time {kpi.value} out of expected range"


"""
Base classes and shared utilities for analytics calculations.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Literal

import pandas as pd

from prodsys.models.production_system_data import ProductionSystemData
from prodsys.models.resource_data import SystemResourceData
from prodsys.simulation import state

import logging

logger = logging.getLogger(__name__)


@dataclass
class AnalyticsContext:
    """
    Context object that holds shared data and configuration for analytics calculations.
    
    This class provides a centralized way to access:
    - Raw simulation data
    - Production system data
    - Configuration (warm-up cutoff, time range, etc.)
    - System resource mappings
    - Sink/source queue names
    """
    
    df_raw: pd.DataFrame
    production_system_data: Optional[ProductionSystemData] = None
    time_range: Optional[float] = None
    warm_up_cutoff: bool = False
    cut_off_method: Optional[Literal["mser5", "threshold_stabilization", "static_ratio"]] = None
    _system_resource_mapping: Optional[dict] = None
    _sink_input_queues: Optional[set] = None
    _source_output_queues: Optional[set] = None
    
    def set_production_system_data(self, production_system_data: ProductionSystemData):
        """Set the production system data and clear manual mappings."""
        self.production_system_data = production_system_data
        self._system_resource_mapping = None
        self._sink_input_queues = None
        self._source_output_queues = None
    
    def set_system_resource_mapping(self, mapping: dict):
        """Set the system resource mapping directly."""
        self._system_resource_mapping = mapping
    
    def set_sink_source_queue_names(self, sink_input_queues: set, source_output_queues: set):
        """Set sink input queue names and source output queue names directly."""
        self._sink_input_queues = sink_input_queues
        self._source_output_queues = source_output_queues
    
    def get_system_resource_mapping(self) -> dict:
        """
        Get mapping of system resource IDs to their subresource IDs.
        
        Returns:
            dict: Mapping from system resource ID to list of subresource IDs.
        """
        if self._system_resource_mapping is not None:
            return self._system_resource_mapping
        
        if self.production_system_data is None:
            return {}
        
        system_resource_mapping = {}
        for resource_data in self.production_system_data.resource_data:
            if isinstance(resource_data, SystemResourceData):
                system_resource_mapping[resource_data.ID] = resource_data.subresource_ids
        
        return system_resource_mapping
    
    def get_sink_source_queue_names(self) -> tuple[set, set]:
        """
        Get names of sink input queues and source output queues.
        
        Returns:
            tuple[set, set]: (sink_input_queues, source_output_queues)
        """
        if self._sink_input_queues is not None and self._source_output_queues is not None:
            return self._sink_input_queues, self._source_output_queues
        
        sink_input_queues = set()
        source_output_queues = set()
        
        if self.production_system_data is None:
            return sink_input_queues, source_output_queues
        
        # Get sink input queues
        for sink_data in self.production_system_data.sink_data:
            if sink_data.ports:
                sink_input_queues.update(sink_data.ports)
        
        # Get source output queues
        for source_data in self.production_system_data.source_data:
            if source_data.ports:
                source_output_queues.update(source_data.ports)
        
        return sink_input_queues, source_output_queues
    
    def get_total_simulation_time(self) -> float:
        """
        Calculates the total simulation time from the data frame.
        
        Returns:
            float: Total simulation time.
        """
        if self.df_raw is not None and "Time" in self.df_raw.columns:
            start_time = self.df_raw["Time"].min()
            end_time = self.df_raw["Time"].max()
            return end_time - start_time
        else:
            raise ValueError("Data frame is not loaded or 'Time' column is missing.")


def get_conditions_for_interface_state(df: pd.DataFrame) -> pd.Series:
    """
    Returns a boolean series indicating whether a row belongs to an interface state.
    
    An interface state is a state where a resource does not perform a process,
    i.e. either setup, breakdown, creation (source) or finish (sink) of products.
    
    Args:
        df: Data frame with the simulation results.
    
    Returns:
        pd.Series: Boolean series indicating interface states.
    """
    return df["State Type"].isin(
        [
            state.StateTypeEnum.source,
            state.StateTypeEnum.sink,
            state.StateTypeEnum.breakdown,
            state.StateTypeEnum.setup,
            state.StateTypeEnum.charging,
            state.StateTypeEnum.loading,
            state.StateTypeEnum.unloading,
            state.StateTypeEnum.assembly,
            state.StateTypeEnum.non_scheduled,
        ]
    )


def get_conditions_for_process_state(df: pd.DataFrame) -> pd.Series:
    """
    Returns a boolean series indicating whether a row belongs to a process state.
    
    A process state is a state where a resource performs a process,
    i.e. either production or transport.
    
    Args:
        df: Data frame with the simulation results.
    
    Returns:
        pd.Series: Boolean series indicating process states.
    """
    return df["State Type"].isin(
        [
            state.StateTypeEnum.production,
            state.StateTypeEnum.transport,
            state.StateTypeEnum.dependency,
        ]
    )




import pytest
from prodsys.models.production_system_data import ProductionSystemData
import prodsys.express as psx
from prodsys import runner


@pytest.fixture
def simulation_adapter() -> ProductionSystemData:
    import prodsys.express as psx

    t1 = psx.FunctionTimeModel("normal", 25, 5, "t1")
    t2 = psx.FunctionTimeModel("normal", 10, 5, "t2")
    t5 = psx.FunctionTimeModel("normal", 5, 0, "t5")
    t6 = psx.FunctionTimeModel("normal", 10, 5, "t6")
    t7 = psx.FunctionTimeModel("normal", 10, 5, "t7")
    t8 = psx.FunctionTimeModel("normal", 60, 8, "t8")
    
    p2 = psx.ProductionProcess(t1, "p2")
    p3 = psx.ProductionProcess(t2, "p3")
    p4 = psx.ProductionProcess(t5, "p4")
    p5 = psx.ProductionProcess(t6, "p5")
    p6 = psx.ProductionProcess(t7, "p6")
    p7 = psx.ProductionProcess(t8, "p7")

    t3 = psx.DistanceTimeModel(speed=5000, reaction_time=0.01, ID="t3")

    tp = psx.TransportProcess(t3, "tp")
    
    move_p = psx.TransportProcess(t3, "move")

    s1 = psx.FunctionTimeModel("exponential", 0.5, ID="s1")

    product1 = psx.Product([p6], tp, "pump",)
    product2 = psx.Product([p3, p5], move_p, "valve block",)


    primitive_dependency1 = psx.AssemblyDependency(
        required_entity= product1,        
    )
    
    primitive_dependency2 = psx.AssemblyDependency(
        required_entity= product2,        
    )
    
    p1 = psx.ProductionProcess(t8, "p1", dependencies= [primitive_dependency1,primitive_dependency2])                          

    machine = psx.Resource(
        [p1],
        [20,10],
        2,        
        ID="assemblymachine_1",
        output_location=[20,10],
        
    )
    
    machine1 = psx.Resource(
        [p2],
        [0,10],
        2,

        ID="machine1",
        output_location=[0,10],
    )
    
    machine2 = psx.Resource(
        [p3],
        [5,0],
        2,
        ID="machine2",
        output_location=[5,0],
    )
    
    machine3 = psx.Resource(
        [p4],
        [15,0],
        2,
        ID="machine3",
        output_location=[15,0],
    )
    
    machine4 = psx.Resource(
        [p5],
        [10,10],
        2,
        ID="machine4",
        output_location=[10,10],
    )
    
    machine5 = psx.Resource(
        [p6],
        [5,10],
        2,
        ID="machine5",
        output_location=[5,10],
    )

    transport = psx.Resource([tp], [2, 2], 3, ID="transport")
    transport2 = psx.Resource([move_p], [2, 2], 3, ID="transporta")
    mutterproduct = psx.Product([p2,p5,p1], tp, "tank")

    sink1 = psx.Sink(product1, [5,20], "sink1")
    sink2 = psx.Sink(product2, [6,10], "sink2")
    sink3 = psx.Sink(mutterproduct, [20.05,10.05], "muttersink")

    arrival_model_1 = psx.FunctionTimeModel("exponential", 22, ID="arrival_model_1")
    arrival_model_2 = psx.FunctionTimeModel("exponential", 22, ID="arrival_model_2")

    source1 = psx.Source(product1, arrival_model_2, [0, 0], ID="source_1")
    source2 = psx.Source(product2, arrival_model_2, [0, 0], ID="source_2")
    source3 = psx.Source(mutterproduct, arrival_model_1, [0, 0], ID="source_3")

    assemblymachine_2 = psx.Resource([p1], [22,10], 2, ID="assemblymachine_2", output_location=[22,10])
     
    resources_A = [machine, machine1, machine2, machine3, machine4, machine5, transport, transport2, assemblymachine_2]
    sources_all = [source1, source2, source3]
    sinks_all = [sink1, sink2, sink3]

    system = psx.ProductionSystem(resources_A, sources_all, sinks_all)
    adapter = system.to_model()
    
    return adapter


def test_initialize_simulation(simulation_adapter: ProductionSystemData):
    runner_instance = runner.Runner(production_system_data=simulation_adapter)
    runner_instance.initialize_simulation()


def test_hashing(simulation_adapter: ProductionSystemData):
    hash_str = simulation_adapter.hash()
    print(hash_str)
    assert hash_str == "a56173e8ae356a5aa4d2b06933c1bf93"


def test_run_simulation(simulation_adapter: ProductionSystemData):
    runner_instance = runner.Runner(production_system_data=simulation_adapter)
    runner_instance.initialize_simulation()
    runner_instance.run(2000)
    runner_instance.print_results()
    assert runner_instance.env.now == 2000
    post_processor = runner_instance.get_post_processor()
    for kpi in post_processor.machine_state_KPIS:
        if kpi.name == "productive_time" and kpi.resource == "assemblymachine_1":
            assert 75 < kpi.value < 85

        if kpi.name == "productive_time" and kpi.resource == "assemblymachine_2":
            assert 75 < kpi.value < 85

    for kpi in post_processor.WIP_KPIs:
        if kpi.name == "WIP" and kpi.product_type == "tank":
            assert kpi.value > 4 and kpi.value < 6

    tank_output = 0.0 
    pump_output = 0.0 
    valve_block_output = 0.0 
    for kpi in post_processor.throughput_and_output_KPIs:
        if kpi.name == "output":
            if kpi.product_type == "tank":
                tank_output = kpi.value
            if kpi.product_type == "valve block":
                valve_block_output = kpi.value
            if kpi.product_type == "pump":
                pump_output = kpi.value
    assert tank_output <= valve_block_output
    assert tank_output <= pump_output
    



from prodsys.models.production_system_data import ProductionSystemData
from prodsys.models.scenario_data import ReconfigurationEnum
from prodsys.optimization.adapter_manipulation import add_transformation_operation
from prodsys.optimization.tabu_search import TabuSearchHyperparameters
from prodsys.optimization.optimizer import FileSystemSaveOptimizer
import logging
import os

logger = logging.getLogger(__name__)


def main():
    hyper_parameters = TabuSearchHyperparameters(
        seed=22,
        tabu_size=2,
        max_steps=5,
        neighborhood_size=5,
        max_score=5000000,
        number_of_seeds=1,
    )

    def new_transformation(adapter: ProductionSystemData) -> bool:
        print("Transformation function called.")

    add_transformation_operation(
        transformation=ReconfigurationEnum.PRODUCTION_CAPACITY,
        operation=new_transformation,
    )

    base_configuration = ProductionSystemData.read(
        "examples/optimization/optimization_example/base_scenario.json",
    )
    base_configuration.read_scenario(
        "examples/optimization/optimization_example/scenario.json",
    )

    initial_solutions = base_configuration
    optimizer = FileSystemSaveOptimizer(
        adapter=base_configuration,
        hyperparameters=hyper_parameters,
        save_folder="data/tabu_results",
        initial_solutions=initial_solutions,
    )

    optimizer.optimize()


if __name__ == "__main__":
    main()



from __future__ import annotations

from enum import Enum
from hashlib import md5
from typing import TYPE_CHECKING, Literal, Optional, Union


from prodsys.models.core_asset import CoreAsset

if TYPE_CHECKING:
    from prodsys.models.production_system_data import ProductionSystemData


class DependencyType(str, Enum):
    """
    Enum class that represents the type of dependency between two components.

    - PROCESS: Dependency on a process.
    - RESOURCE: Dependency on a resource.
    - TOOL: Dependency on a tool. For example, a tool needed for a production process (e.g. a drill bit) or a work piece carrier for a transport process.
    - ASSEMBLY: Dependency on an assembly. A primitive or product needed for an assembly process.
    - DISASSEMBLY: Dependency on a disassembly. A primitive or product created by an disassembly process.
    - LOT: Dependency on a lot. For example, a lot needed for a production process (e.g. a lot of screws) or a work piece carrier for a transport process.
    """
    PROCESS = "process"
    RESOURCE = "resource"
    TOOL = "tool"
    ASSEMBLY = "assembly"
    DISASSEMBLY = "disassembly"
    LOT = "lot"

class DependencyData(CoreAsset):
    dependency_type: DependencyType


class ProcessDependencyData(DependencyData):
    required_process: str
    dependency_type: DependencyType = DependencyType.PROCESS
    interaction_node: Optional[str] = None
    position_type: Literal["relative", "absolute"] = "absolute"
    per_lot: bool = True

    def hash(self, adapter: ProductionSystemData) -> str:
        """
        Function to hash the process dependency.

        Returns:
            str: Hash of the process dependency.
        """
        process_hash = ""
        for process in adapter.process_data:
            if process.ID == self.required_process:
                process_hash = process.hash(adapter)
        return md5(
            "".join([self.dependency_type, process_hash]).encode("utf-8")
        ).hexdigest()


class ResourceDependencyData(DependencyData):
    required_resource: str
    dependency_type: DependencyType = DependencyType.RESOURCE
    interaction_node: Optional[str] = None
    position_type: Literal["relative", "absolute"] = "absolute"
    per_lot: bool = True
    
    def hash(self, adapter: ProductionSystemData) -> str:
        """
        Function to hash the resource dependency.

        Returns:
            str: Hash of the resource dependency.
        """
        resource_hash = ""
        for resource in adapter.resource_data:
            if resource.ID == self.required_resource:
                resource_hash = resource.hash(adapter)
        return md5(
            "".join([self.dependency_type, resource_hash]).encode("utf-8")
        ).hexdigest()


class ToolDependencyData(DependencyData):
    """
    Class that defines a dependency on a tool. This could mean, e.g. that a product requires a workpiece carrier to be transported or a tool to be processed. The tool is not a product itself, but a component that is needed for the production process.

    Args:
        ID (str): ID of the dependency.
        description (str): Description of the dependency.
        required_primitive (str): ID of the required primitive.
        dependency_type (DependencyType, optional): Type of the dependency. Defaults to DependencyType.PRIMITIVE.
        per_lot (bool, optional): If True, the dependency is needed once per lot. If False, it's needed for every instance in the lot. Defaults to False.

    Returns:
        _type_: _description_
    """

    required_entity: str
    dependency_type: DependencyType = DependencyType.TOOL
    per_lot: bool = True

    def hash(self, adapter: ProductionSystemData) -> str:
        """
        Function to hash the primitive dependency.

        Returns:
            str: Hash of the primitive dependency.
        """
        entity_hash = ""
        for entity in adapter.primitive_data + adapter.product_data:
            if entity.ID == self.required_entity:
                entity_hash = entity.hash(adapter)
        return md5(
            "".join([self.dependency_type, entity_hash]).encode("utf-8")
        ).hexdigest()


class AssemblyDependencyData(DependencyData):
    required_entity: str
    dependency_type: DependencyType = DependencyType.ASSEMBLY
    per_lot: bool = False

    def hash(self, adapter: ProductionSystemData) -> str:
        """
        Function to hash the assembly dependency.

        Returns:
            str: Hash of the assembly dependency.
        """
        entity_hash = ""
        for entity in adapter.primitive_data + adapter.product_data:
            if entity.ID == self.required_entity:
                entity_hash = entity.hash(adapter)
        return md5(
            "".join([self.dependency_type, entity_hash]).encode("utf-8")
        ).hexdigest()


class DisassemblyDependencyData(DependencyData):
    required_entity: str
    dependency_type: DependencyType = DependencyType.DISASSEMBLY

    def hash(self, adapter: ProductionSystemData) -> str:
        """
        Function to hash the disassembly dependency.

        Returns:
            str: Hash of the disassembly dependency.
        """
        entity_hash = ""
        for entity in adapter.primitive_data + adapter.product_data:
            if entity.ID == self.required_entity:
                entity_hash = entity.hash(adapter)
        return md5(
            "".join([self.dependency_type, entity_hash]).encode("utf-8")
        ).hexdigest()


class LotDependencyData(DependencyData):
    """
    Class that defines that processes need to be performed with a carrier, specifying also how many (max / min) products should be placed on this carrier
    """
    min_lot_size: int = 1
    max_lot_size: int = 1
    input_output: Literal["input", "output", "input_output"] = "input_output"
    dependency_type: DependencyType = DependencyType.LOT

    def hash(self, adapter: ProductionSystemData) -> str:
        """
        Function to hash the lot dependency.

        Returns:
            str: Hash of the lot dependency.
        """
        return md5(
            "".join([str(self.dependency_type.value), str(self.min_lot_size), str(self.max_lot_size), str(self.input_output)]).encode("utf-8")
        ).hexdigest()


DEPENDENCY_TYPES = Union[
    ProcessDependencyData,
    ResourceDependencyData,
    ToolDependencyData,
    AssemblyDependencyData,
    DisassemblyDependencyData,
    LotDependencyData,
]


from __future__ import annotations

from typing import TYPE_CHECKING, List, Generator, Optional, Union

import random

import logging

from prodsys.models import port_data
from prodsys.simulation.dependency import DependencyInfo



from simpy.resources import resource
from simpy import events

if TYPE_CHECKING:
    from prodsys.simulation import control, state
    from prodsys.simulation.dependency import DependedEntity, Dependency

    # from prodsys.simulation.process import PROCESS_UNION
    from prodsys.simulation import sim

    from prodsys.simulation.process import PROCESS_UNION
    from prodsys.simulation.locatable import Locatable
    from prodsys.simulation import router


from prodsys.models.resource_data import (
    ResourceData,
    SystemResourceData,
)
from prodsys.util import util

logger = logging.getLogger(__name__)


class Resource(resource.Resource):
    """
    Base class for all resources.

    Args:
        env (sim.Environment): The simpy environment.
        data (RESOURCE_DATA_UNION): The resource data.
        processes (List[PROCESS_UNION]): The processes.
        controller (control.Controller): The controller.
        states (List[state.State]): The states of the resource for breakdowns.
        production_states (List[state.State]): The states of the resource for production.
        setup_states (List[state.SetupState]): The states of the resource for setups.
        active (events.Event): The event that is triggered when the resource is active.
        current_setup (PROCESS_UNION): The current setup.
        reserved_setup (PROCESS_UNION): The reserved setup.
    """

    def __init__(
        self,
        env: sim.Environment,
        data: ResourceData,
        processes: List[PROCESS_UNION],
        controller: control.Controller,
        can_move: bool = False,
        can_process: bool = False,
        states: List[state.State] = None,
        production_states: List[state.State] = None,
        setup_states: List[state.SetupState] = None,
        charging_states: List[state.ChargingState] = None,
        ports: List[port.Queue] = None,
        buffers: List[port.Queue] = None
    ):
        super().__init__(env, capacity=data.capacity)
        self.env = env
        self.data = data
        self.processes = processes
        self.controller = controller
        self.states = states if states else []
        self.production_states = production_states if production_states else []
        self.setup_states = setup_states if setup_states else []
        self.charging_states = charging_states if charging_states else []

        self.dependencies: List[Dependency] = []
        self.depended_entities: List[DependedEntity] = []
        self.dependency_info: DependencyInfo = DependencyInfo(resource_id=data.ID)

        self.active = events.Event(self.env).succeed()
        self.current_setup: PROCESS_UNION = None
        self.reserved_setup: PROCESS_UNION = None

        self.can_move = can_move
        self.can_process = can_process

        self.bound = False
        self.current_dependant: Union[Resource] = None

        self.ports = ports if ports else []
        self.buffers = buffers if buffers else []
        self.current_locatable = self

        self.full = False

    def bind_to_dependant(self, dependant: Resource) -> None:
        """
        Binds the resource to a depended entity.

        Args:
            dependant (DependedEntity): The depended entity.
        """
        self.current_dependant = dependant
        self.bound = True

    def release_from_dependant(self) -> None:
        """
        Releases the resource from the depended entity.
        """
        self.current_dependant = None
        self.bound = False

    @property
    def capacity_current_setup(self) -> int:
        """
        Returns the capacity of the resource for the current setup with considering that the resource could be in a setup process.

        Returns:
            int: The capacity of the resource for the current setup.
        """
        if not self.current_setup and not self.reserved_setup:
            return self.capacity
        
        # Determine which setup to use for capacity calculation
        setup_to_check = None
        if (
            self.reserved_setup
            and self.current_setup
            and self.current_setup.data.ID != self.reserved_setup.data.ID
        ):
            setup_to_check = self.reserved_setup
        elif self.reserved_setup:
            setup_to_check = self.reserved_setup
        elif self.current_setup:
            setup_to_check = self.current_setup
        
        # If setup is a ProcessModelProcess, it doesn't have production states
        # so return base capacity instead
        from prodsys.simulation.process import ProcessModelProcess
        if setup_to_check and isinstance(setup_to_check, ProcessModelProcess):
            return self.capacity
        
        current_setup_ID = setup_to_check.data.ID
        length = len(
            [
                state
                for state in self.production_states
                if state.data.ID == current_setup_ID
            ]
        )
        # If no production states found for the setup, fall back to base capacity
        # This can happen for process models or other processes without explicit production states
        return length if length > 0 else self.capacity

    def wait_for_free_process(
        self, process: PROCESS_UNION
    ) -> Generator[state.State, None, None]:
        """
        Wait for a free process of a resource.

        Args:
            resource (resources.ResourceData): The resource.
            process (process.Process): The process.

        Returns:
            Generator: The generator yields when a process is free.

        Yields:
            Generator: The generator yields when a process is free.
        """
        while True:
            free_state = self.get_free_process(process)
            if free_state is not None:
                return free_state
            yield self.controller.state_changed

    def reserve_setup(self, process: PROCESS_UNION) -> None:
        """
        Reserves the setup of the resource for a process. This is used to prevent that capacity is wrong estimated during setup.

        Args:
            process (PROCESS_UNION): The process that wants to reserve the setup.
        """
        self.reserved_setup = process

    def unreserve_setup(self) -> None:
        """
        Unreserves the setup of the resource. This is used to prevent that the resource is used for another process while it is in a setup process.
        """
        self.reserved_setup = None

    @property
    def in_setup(self) -> bool:
        """
        Returns if the resource is in a setup process.

        Returns:
            bool: True if the resource is in a setup process, False otherwise.
        """
        return self.reserved_setup is not None

    def update_full(self) -> None:
        """
        Returns if the resource is full.

        Returns:
            bool: True if the resource is full or in setup, False otherwise.
        """
        self.full = self.get_free_capacity() <= 0

    def get_free_capacity(self) -> int:
        """
        Returns the free capacity of the resource.

        Returns:
            int: The free capacity of the resource.
        """
        return self.capacity_current_setup - (
            self.controller.num_running_processes
            + self.controller.reserved_requests_count
        )

    @property
    def requires_charging(self) -> bool:
        """
        Returns if the resource requires charging.

        Returns:
            bool: True if the resource requires charging, False otherwise.
        """
        return any(
            [
                state_instance.requires_charging()
                for state_instance in self.charging_states
                if isinstance(state_instance, state.ChargingState)
            ]
        )

    def charge(self) -> Generator:
        """
        Charges the resource.

        Yields:
            Generator: The type of the yield depends on the resource.
        """
        for input_state in self.charging_states:
            if not input_state.requires_charging():
                continue
            yield self.env.process(input_state.process_state())

    def consider_battery_usage(self, amount: float) -> None:
        """
        Reduces the battery level of the resource.

        Args:
            amount (float): The amount to reduce the battery level.
        """
        for state_instance in self.charging_states:
            state_instance.add_battery_usage_time(amount)

    def get_controller(self) -> control.Controller:
        """
        Returns the controller of the resource.

        Returns:
            control.Controller: The controller of the resource.
        """
        return self.controller

    def add_state(self, input_state: state.STATE_UNION) -> None:
        """
        Adds a state to the resource.

        Args:
            input_state (state.STATE_UNION): The state to add.
        """
        if isinstance(input_state, state.SetupState):
            self.setup_states.append(input_state)
        elif isinstance(input_state, state.ChargingState):
            self.charging_states.append(input_state)
        else:
            self.states.append(input_state)
        input_state.set_resource(self)

    def add_production_state(self, input_state: state.ProductionState) -> None:
        """
        Adds a production state to the resource.

        Args:
            input_state (state.ProductionState): The production state to add.
        """
        self.production_states.append(input_state)
        input_state.set_resource(self)

    def start_states(self):
        """
        Starts the simpy processes of the states of the resource in simpy.
        """
        resource.Resource.__init__(self, self.env, capacity=self.data.capacity)
        self.active = events.Event(self.env).succeed()
        for actual_state in (
            self.states
            + self.production_states
            + self.setup_states
            + self.charging_states
        ):
            actual_state.activate_state()
        for actual_state in self.states:
            actual_state.process = self.env.process(actual_state.process_state())

    def get_process(self, process: PROCESS_UNION) -> state.State:
        """
        Returns the ProducitonState or CapabilityState of the resource for a process.

        Args:
            process (PROCESS_UNION): The process to get the state for.

        Raises:
            ValueError: If the process is not found in the resource.

        Returns:
            state.State: The state of the resource for the process.
        """
        possible_states = [
            actual_state
            for actual_state in self.production_states
            if actual_state.data.ID == process.data.ID
        ]
        if not possible_states:
            raise ValueError(
                f"Process {process.data.ID} not found in resource {self.data.ID}"
            )
        return random.choice(possible_states)

    def get_processes(self, process: PROCESS_UNION) -> List[state.State]:
        """
        Returns the ProducitonState or CapabilityState of the resource for a process.

        Args:
            process (PROCESS_UNION): The process to get the state for.

        Raises:
            ValueError: If the process is not found in the resource.

        Returns:
            List[state.State]: The state of the resource for the process.
        """
        possible_states = [
            actual_state
            for actual_state in self.production_states
            if actual_state.data.ID == process.data.ID
        ]
        if not possible_states:
            raise ValueError(
                f"Process {process.data.ID} not found in resource {self.data.ID}"
            )
        return possible_states

    def get_free_process(self, process: PROCESS_UNION) -> Optional[state.State]:
        """
        Returns a free ProductionState or CapabilityState of the resource for a process.

        Args:
            process (PROCESS_UNION): The process to get the state for.

        Returns:
            Optional[state.State]: The state of the resource for the process.
        """
        for actual_state in self.production_states:
            if (
                actual_state.data.ID == process.data.ID
                and actual_state.process is None
                and not actual_state.reserved
            ):
                return actual_state
        return None

    def get_free_processes(self, process: PROCESS_UNION) -> Optional[List[state.State]]:
        """
        Returns all free ProductionState or CapabilityState of the resource for a process.

        Args:
            process (process.PROCESS_UNION): The process to get the state for.

        Returns:
            List[state.State]: The state of the resource for the process.
        """
        return [
            actual_state
            for actual_state in self.production_states
            if actual_state.data.ID == process.data.ID
            and (actual_state.process is None or not actual_state.process.is_alive)
        ]

    def get_location(
        self
    ) -> List[float]:
        return self.data.location

    def get_input_queue_length(self) -> int:
        """
        Returns total number of items in all input_queues.

        Returns:
            int: Sum of items in the resources input-queues.
        """
        return sum([len(q.items) for q in self.ports if q.data.interface_type in [port_data.PortInterfaceType.INPUT, port_data.PortInterfaceType.INPUT_OUTPUT]])

    def get_output_queue_length(self) -> int:
        """
        Returns total number of items in all output_queues.

        Returns:
            int: Sum of items in the resources output-queues.
        """
        return sum([len(q.items) for q in self.ports if q.data.interface_type in [port_data.PortInterfaceType.OUTPUT, port_data.PortInterfaceType.INPUT_OUTPUT]])

    def set_location(self, new_location: Locatable) -> None:
        """
        Sets the location of the resource.

        Args:
            new_location (List[float]): The new location of the resource. Has to have length 2.
        """
        self.data.location = new_location.get_location()
        self.current_locatable = new_location

    def get_states(self) -> List[state.State]:
        """
        Returns the states of the resource.

        Returns:
            List[state.State]: The states of the resource.
        """
        return self.states

    def activate(self):
        """
        Activates the resource after a breakdwon or non-scheduled period.
        """
        if any(
            [
                state_instance.active_breakdown
                for state_instance in self.states
                if isinstance(state_instance, state.BreakDownState)
            ]
        ):
            return
        if any(
            [
                state_instance.active_non_scheduled
                for state_instance in self.states
                if isinstance(state_instance, state.NonScheduledState)
            ]
        ):
            return
        self.active.succeed()

    def interrupt_states(self):
        """
        Interrupts the states of the resource.
        """
        if self.active.triggered:
            self.active = events.Event(self.env)
        for state_instance in self.setup_states + self.production_states:
            if (
                state_instance.process
                and state_instance.process.is_alive
                and not state_instance.interrupted
            ):
                state_instance.interrupt_process()

    def get_free_of_setups(self) -> Generator:
        """
        Returns a generator that yields when all setups are finished.

        Yields:
            Generator: The generator of the yield, which is yielded when all setups are finished.
        """
        running_setups = [
            state.process
            for state in self.setup_states
            if (state.process and state.process.is_alive)
        ]
        yield events.AllOf(self.env, running_setups)

    def setup(self, _process: PROCESS_UNION) -> Generator:
        """
        Sets up the resource for a process.

        Args:
            _process (PROCESS_UNION): The process to set up the resource for.

        Yields:
            Generator: The type of the yield depends on the process.
        """
        if self.current_setup is None:
            yield self.env.process(util.trivial_process(self.env))
            self.current_setup = _process
            return
        if self.reserved_setup:
            setup_to_compare = self.reserved_setup
        else:
            setup_to_compare = self.current_setup

        if setup_to_compare.data.ID == _process.data.ID:
            yield self.env.process(self.get_free_of_setups())
            yield self.env.process(util.trivial_process(self.env))
            return

        for input_state in self.setup_states:
            if (
                input_state.data.target_setup == _process.data.ID
                and input_state.data.origin_setup == setup_to_compare.data.ID
            ):
                self.reserve_setup(_process)
                yield self.env.process(self.get_free_of_setups())
                input_state.process = self.env.process(input_state.process_state())
                yield input_state.process
                input_state.process = None
                self.current_setup = _process
                self.unreserve_setup()

        else:
            yield self.env.process(self.get_free_of_setups())
            yield self.env.process(util.trivial_process(self.env))

    def add_ports(self, ports: List[port.Queue]):
        self.ports.extend(ports)


class SystemResource(Resource):
    """
    Class that represents a system resource. A system resource is a resource with subresources that can be used interchangeably.

    Args:
        env (sim.Environment): The simpy environment.
        data (SystemResourceData): The system resource data.
        processes (List[PROCESS_UNION]): The processes.
        controller (control.Controller): The controller.
        subresources (List[Resource]): The subresources of this system.
        system_ports (List[port.Queue]): The system ports for external communication.
        internal_routing_matrix (Dict[str, List[str]]): Internal routing matrix for routing within the system.
        can_move (bool): Whether the resource can move.
        can_process (bool): Whether the resource can process.
        states (List[state.State]): The states of the resource for breakdowns.
        production_states (List[state.State]): The states of the resource for production.
        setup_states (List[state.SetupState]): The states of the resource for setups.
        charging_states (List[state.ChargingState]): The states of the resource for charging.
        ports (List[port.Queue]): The ports of the resource.
    """

    def __init__(
        self,
        env: sim.Environment,
        data: SystemResourceData,
        processes: List[PROCESS_UNION],
        controller: control.Controller,
        subresources: List[Resource],
        can_move: bool = False,
        can_process: bool = False,
        states: List[state.State] = None,
        production_states: List[state.State] = None,
        setup_states: List[state.SetupState] = None,
        charging_states: List[state.ChargingState] = None,
        ports: List[port.Queue] = None,
        buffers: List[port.Queue] = None
    ):
        # Store original capacity for systems with capacity 0
        self._actual_capacity = data.capacity
        
        # Just to avoid an instantiation error for simpy resources, which does not allow capacity to be 0
        if data.capacity == 0:
            data.capacity = 1
            
        super().__init__(
            env, data, processes, controller, can_move, can_process,
            states, production_states, setup_states, charging_states, ports, buffers
        )

        self.subresources = subresources
        self.router: router.Router = None        
        # Create mapping for quick lookup
        self.subresource_map = {resource.data.ID: resource for resource in self.subresources}
        self.system_port_map = {port.data.ID: port for port in self.ports}

    def set_router(self, router: router.Router) -> None:
        """
        Sets the router of the system resource.
        """
        self.router = router

    def update_full(self) -> None:
        """
        Returns if the resource is full.

        Returns:
            bool: True if the resource is full or in setup, False otherwise.
        """
        if self._actual_capacity == 0:
            self.full = False
            return
        self.full = self.get_free_capacity() <= 0

    def get_free_capacity(self) -> int:
        """
        Returns the free capacity of the resource.

        Returns:
            int: The free capacity of the resource.
        """
        if self._actual_capacity == 0:
            return float("inf")
        return self._actual_capacity - (
            self.controller.num_running_processes
            + self.controller.reserved_requests_count
        )

    def get_system_port(self, port_id: str) -> Optional[port.Queue]:
        """
        Get a system port by ID.

        Args:
            port_id (str): The port ID.

        Returns:
            Optional[port.Queue]: The system port if found, None otherwise.
        """
        return self.system_port_map.get(port_id)

    def get_subresource(self, resource_id: str) -> Optional[Resource]:
        """
        Get a subresource by ID.

        Args:
            resource_id (str): The resource ID.

        Returns:
            Optional[Resource]: The subresource if found, None otherwise.
        """
        return self.subresource_map.get(resource_id)

    def can_handle_process(self, process_id: str) -> bool:
        """
        Check if this system resource can handle the given process.

        Args:
            process_id (str): The process ID to check.

        Returns:
            bool: True if the system can handle the process.
        """
        # Check if any subresource can handle the process
        return any(
            any(process.data.ID == process_id for process in subresource.processes)
            for subresource in self.subresources
        )


RESOURCE_UNION = Union[Resource, SystemResource]
""" Union Type for Resources. """

from prodsys.simulation import port, state



from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum
from typing import Literal, Optional, Union, TYPE_CHECKING, Generator, List

import logging


from simpy import events
from simpy import exceptions

from prodsys.models.state_data import (
    StateData,
    BreakDownStateData,
    ProductionStateData,
    TransportStateData,
    SetupStateData,
    ProcessBreakDownStateData,
    ChargingStateData,
    NonScheduledStateData,
)

if TYPE_CHECKING:
    from prodsys.simulation import resources
    from prodsys.simulation import sim, time_model
    from prodsys.simulation.entities import product, primitive
    from prodsys.simulation.locatable import Locatable

logger = logging.getLogger(__name__)


class StateEnum(str, Enum):
    """
    Enum for the different types a state can be in.
    """

    start_state = "start state"
    start_interrupt = "start interrupt"
    end_interrupt = "end interrupt"
    end_state = "end state"

    created_product = "created product"
    consumed_product = "consumed product"
    finished_product = "finished product"

    created_primitive = "created primitive"
    started_primitive_usage = "started primitive usage"
    finished_primitive_usage = "finished primitive usage"


class StateTypeEnum(str, Enum):
    """
    Enum for the different types of states.
    """

    production = "Production"
    transport = "Transport"
    breakdown = "Breakdown"
    process_breakdown = "ProcessBreakdown"
    setup = "Setup"
    source = "Source"
    assembly = "Assembly"
    disassembly = "Disassembly"
    sink = "Sink"
    store = "Store"
    charging = "Charging"
    dependency = "Dependency"
    loading = "Loading"
    unloading = "Unloading"
    non_scheduled = "NonScheduled"


class StateInfo:
    """
    Class that represents the current event information of a state while simulating.

    Args:
        ID (str): The ID of the state.
        resource_ID (str): The ID of the resource the state belongs to.
        _event_time (Optional[float], optional): The time of the event. Defaults to 0.0.
        _expected_end_time (Optional[float], optional): The expected end time of the state. Defaults to 0.0.
        _activity (Optional[StateEnum], optional): The activity of the state. Defaults to None.
        _state_type (Optional[StateTypeEnum], optional): The type of the state. Defaults to None.
        _product_ID (str, optional): The ID of the product the state belongs to. Defaults to "".
        _target_ID (str, optional): The ID of the target the state belongs to. Defaults to "".
    """

    def __init__(
        self,
        ID: str,
        resource_ID: str,
        _event_time: Optional[float] = 0.0,
        _expected_end_time: Optional[float] = 0.0,
        _activity: Optional[StateEnum] = None,
        _state_type: Optional[StateTypeEnum] = None,
        _product_ID: str = "",
        _target_ID: str = "",
        _origin_ID: str = "",
        _dependency_ID: str = "",
        _empty_transport: Optional[bool] = None,
        _initial_transport_step: bool = False,
        _last_transport_step: bool = False,
    ):
        self.ID = ID
        self.resource_ID = resource_ID
        self._event_time = _event_time
        self._expected_end_time = _expected_end_time
        self._activity = _activity
        self._state_type = _state_type
        self._product_ID = _product_ID
        self._target_ID = _target_ID
        self._origin_ID = _origin_ID
        self._dependency_ID = _dependency_ID
        self._empty_transport = _empty_transport
        self._initial_transport_step = _initial_transport_step
        self._last_transport_step = _last_transport_step
        self._process_ok: bool = True  # Track if process was successful (True) or failed (False)
        
    def log_transport(
        self,
        origin: Optional[Locatable],
        target: Locatable,
        state_type: StateTypeEnum,
        empty_transport: bool,
        initial_transport_step: bool,
        last_transport_step: bool,
    ):
        """
        Logs the target location of a transport state.

        Args:
            origin (Optional[Locatable]): The origin location, either a resource, source, node or a sink.
            target (Locatable): The target location, either a resource, source, node or a sink.
            state_type (StateTypeEnum): The type of the state.
            empty_transport (bool): Indicates if the transport is empty.
        """
        if not origin:
            self._origin_ID = "Loading station"
        else:
            self._origin_ID = origin.data.ID
        self._target_ID = target.data.ID
        self._state_type = state_type
        self._empty_transport = empty_transport
        self._initial_transport_step = initial_transport_step
        self._last_transport_step = last_transport_step

    def log_product(self, _product: product.Product, state_type: StateTypeEnum):
        """
        Logs the product of a transport or production state.

        Args:
            _product (product.Product): The product.
            state_type (StateTypeEnum): The type of the state.
        """
        self._product_ID = _product.data.ID
        self._state_type = state_type

    def log_primitive(self, _primitive: primitive.Primitive, state_type: StateTypeEnum):
        """
        Logs the product of a transport or production state.

        Args:
            _product (product.Product): The product.
            state_type (StateTypeEnum): The type of the state.
        """
        self._product_ID = _primitive.data.ID
        self._state_type = state_type

    def log_start_state(
        self, start_time: float, expected_end_time: float, state_type: StateTypeEnum
    ):
        """
        Logs the start of a state.

        Args:
            start_time (float): The start time of the state.
            expected_end_time (float): The expected end time of the state.
            state_type (StateTypeEnum): The type of the state.
        """
        self._event_time = start_time
        self._expected_end_time = expected_end_time
        self._activity = StateEnum.start_state
        self._state_type = state_type

    def log_start_interrupt_state(self, start_time: float, state_type: StateTypeEnum):
        """
        Logs the start of an interrupt of a state.

        Args:
            start_time (float): The start time of the interruption.
            state_type (StateTypeEnum): The type of the state.
        """
        self._event_time = start_time
        self._activity = StateEnum.start_interrupt
        self._state_type = state_type

    def log_end_interrupt_state(
        self, end_time: float, expected_end_time: float, state_type: StateTypeEnum
    ):
        """
        Logs the end of an interrupt of a state.

        Args:
            end_time (float): The end time of the interruption.
            expected_end_time (float): The expected end time of the state.
            state_type (StateTypeEnum): The type of the state.
        """
        self._event_time = end_time
        self._expected_end_time = expected_end_time
        self._activity = StateEnum.end_interrupt
        self._state_type = state_type

    def log_end_state(self, end_time: float, state_type: StateTypeEnum, process_ok: bool = True):
        """
        Logs the end of a state.

        Args:
            end_time (float): The end time of the state.
            state_type (StateTypeEnum): The type of the state.
            process_ok (bool): Whether the process was successful. Defaults to True.
        """
        self._event_time = end_time
        self._activity = StateEnum.end_state
        self._state_type = state_type
        self._process_ok = process_ok


class State(ABC):
    """
    Abstract class that represents a state of a resource in the simulation. A state has a process that is simulated when the resource starts a state. States can exist in parallel and can interrupt each other.

    Args:
        data (StateData): The data of the state.
        time_model (time_model.TimeModel): The time model of the state.
        env (sim.Environment): The simulation environment.
        active (events.Event, optional): Event that indidcates if the state is active. Defaults to None.
        resource (resources.Resource, optional): The resource the state belongs to. Defaults to None.
        process (Optional[events.Process], optional): The process of the state. Defaults to None.
        state_info (StateInfo, optional): The state information of the state. Defaults to None.
    """

    def __init__(
        self,
        data: StateData,
        time_model: time_model.TimeModel,
        env: sim.Environment,
    ):
        self.data = data
        self.time_model = time_model
        self.env = env
        self.process: Optional[events.Process] = None
        self.active = events.Event(self.env)
        self.resource = None
        self.reserved = False

    def set_resource(self, resource_model: resources.Resource) -> None:
        """
        Sets the resource of the state.

        Args:
            resource_model (resources.Resource): The resource the state belongs to.
        """
        self.resource = resource_model
        self.state_info = StateInfo(ID=self.data.ID, resource_ID=self.resource.data.ID)

    def activate_state(self):
        self.active = events.Event(self.env).succeed()

    def deactivate(self):
        """
        Deactivates the state by setting the active event to a new event which is not yet triggered.
        """
        self.active = events.Event(self.env)

    def activate(self):
        """
        Activates the state by triggering the active event.

        Raises:
            RuntimeError: If the state is allready active.
        """
        try:
            self.active.succeed()
        except RuntimeError:
            raise RuntimeError(f"state {self.data.ID} is allready succeded!!")

    @abstractmethod
    def process_state(self) -> Generator:
        """
        Runs a single process of the state. The process is the key component for the behavior of the state while simulating. All the logic of the state is implemented in the process.

        Yields:
            Generator: The generator of the process.
        """
        pass

    @abstractmethod
    def interrupt_process(self):
        """
        Interrupts the process of the state.
        """
        pass


class ProductionState(State):
    """
    Represents a production state of a resource in the simulation. A production state has a process that simulates the production process which takes some time. The production state continues the creation process of a product. If a resource has a higher capacity than 1 for a process, multiple production states exist, that can run in parallel.

    Args:
        data (ProductionStateData): The data of the state.
        time_model (time_model.TimeModel): The time model of the state.
        env (sim.Environment): The simulation environment.
        active (events.Event, optional): Event that indidcates if the state is active. Defaults to None.
        resource (resources.Resource, optional): The resource the state belongs to. Defaults to None.
        process (Optional[events.Process], optional): The process of the state. Defaults to None.
        state_info (StateInfo, optional): The state information of the state. Defaults to None.
        start (float, optional): The start time of the state. Defaults to 0.0.
        done_in (float, optional): The ramaining time for the state to finish. Defaults to 0.0.
        interrupted (bool, optional): Indicates if the state is interrupted. Defaults to False.
    """

    def __init__(
        self,
        data: ProductionStateData,
        time_model: time_model.TimeModel,
        env: sim.Environment,
    ):
        super().__init__(data, time_model, env)
        self.start = 0.0
        self.done_in = 0.0
        self.interrupted = False

    def process_state(self, time: Optional[float] = None) -> Generator:
        if not time:
            time = self.time_model.get_next_time()
        self.done_in = time
        self.resource.consider_battery_usage(self.done_in)
        
        # Determine if the process will fail BEFORE execution
        process_failed = self._determine_process_failure()
        
        while True:
            try:
                if self.interrupted:
                    yield events.AllOf(self.env, [self.active, self.resource.active])
                    self.interrupted = False
                yield events.AllOf(self.env, [self.resource.active, self.active])
                break
            except exceptions.Interrupt:
                if not self.interrupted:
                    raise RuntimeError(
                        f"Simpy interrupt occured at {self.data.ID} although process is not interrupted"
                    )
        while self.done_in:
            try:
                if self.interrupted:
                    self.state_info.log_start_interrupt_state(
                        self.env.now, StateTypeEnum.production
                    )
                    self.update_done_in()
                    yield events.AllOf(self.env, [self.active, self.resource.active])
                    self.interrupted = False
                    self.state_info.log_end_interrupt_state(
                        self.env.now,
                        self.env.now + self.done_in,
                        StateTypeEnum.production,
                    )
                self.start = self.env.now
                self.state_info.log_start_state(
                    self.start, self.start + self.done_in, StateTypeEnum.production
                )
                yield self.env.timeout(self.done_in)
                self.done_in = 0  # Set to 0 to exit while loop.

            except exceptions.Interrupt:
                if not self.interrupted:
                    raise RuntimeError(
                        f"Simpy interrupt occured at {self.data.ID} although process is not interrupted"
                    )
        # Log end state with success/failure status (process_ok is True if not failed, False if failed)
        process_ok = not process_failed if process_failed is not None else True
        self.state_info.log_end_state(self.env.now, StateTypeEnum.production, process_ok=process_ok)
    
    def _determine_process_failure(self) -> Optional[bool]:
        """
        Determine if the process will fail based on the state's failure_rate.
        
        Returns:
            Optional[bool]: True if the process will fail, False if it will succeed, None if failure_rate is not applicable.
        """
        if not hasattr(self.data, 'failure_rate'):
            return None
            
        failure_rate = self.data.failure_rate
        if not failure_rate or failure_rate == 0:
            return False
        
        import numpy as np  # noqa: E402
        will_fail = bool(np.random.choice(
            [True, False], p=[failure_rate, 1 - failure_rate]
        ))
        return will_fail

    def update_done_in(self):
        if self.start == 0:
            return
        self.done_in -= self.env.now - self.start  # How much time left?
        if self.done_in < 0:
            self.done_in = 0

    def interrupt_process(self):
        if self.process and self.process.is_alive and not self.interrupted:
            self.interrupted = True
            self.process.interrupt()


class TransportState(State):
    """
    Represents a transport state of a resource in the simulation. A transport state has a process that simulates the transport of a product. The transport state continues the transport process of a product. If a resource has a higher capacity than 1 for a process, multiple transport states exist, that can run in parallel but only with the same target and end location.

    Args:
        data (TransportStateData): The data of the state.
        time_model (time_model.TimeModel): The time model of the state.
        env (sim.Environment): The simulation environment.
        active (events.Event, optional): Event that indidcates if the state is active. Defaults to None.
        resource (resources.Resource, optional): The resource the state belongs to. Defaults to None.
        process (Optional[events.Process], optional): The process of the state. Defaults to None.
        state_info (StateInfo, optional): The state information of the state. Defaults to None.
        start (float, optional): The start time of the state. Defaults to 0.0.
        done_in (float, optional): The ramaining time for the state to finish. Defaults to 0.0.
        interrupted (bool, optional): Indicates if the state is interrupted. Defaults to False.
        loading_time_model (time_model.TimeModel, optional): The time model of the loading time. Defaults to None.
    """

    def __init__(
        self,
        data: TransportStateData,
        time_model: time_model.TimeModel,
        env: sim.Environment,
        loading_time_model: Optional[time_model.TimeModel] = None,
        unloading_time_model: Optional[time_model.TimeModel] = None,
    ):
        super().__init__(data, time_model, env)
        self.loading_time_model = loading_time_model
        self.unloading_time_model = unloading_time_model
        self.start = 0.0
        self.done_in = 0.0
        self.interrupted = False
        self.loading_time = 0.0
        self.unloading_time = 0.0

    def get_handling_time(self, action: Literal["loading, unloading"]) -> float:
        if action == "loading":
            time_model = self.loading_time_model
        elif action == "unloading":
            time_model = self.unloading_time_model
        else:
            raise ValueError(f"Unknown action {action}")

        return time_model.get_next_time() if time_model else 0

    def process_state(
        self,
        target: List[float],
        empty_transport: bool,
        initial_transport_step: bool,
        last_transport_step: bool,
    ) -> Generator:
        self.done_in = self.time_model.get_next_time(
            origin=self.resource.get_location(), target=target
        )
        if (
            initial_transport_step
            and hasattr(self.time_model, "reaction_time")
            and self.time_model.data.reaction_time
        ):
            self.done_in += self.time_model.data.reaction_time
        if self.loading_time_model and initial_transport_step and not empty_transport:
            self.loading_time = self.get_handling_time("loading")
            self.done_in += self.loading_time
        if self.unloading_time_model and last_transport_step and not empty_transport:
            self.unloading_time = self.get_handling_time("unloading")
            self.done_in += self.unloading_time
        self.resource.consider_battery_usage(self.done_in)

        while True:
            try:
                if self.interrupted:
                    yield events.AllOf(self.env, [self.active, self.resource.active])
                    self.interrupted = False
                yield events.AllOf(self.env, [self.resource.active, self.active])
                break
            except exceptions.Interrupt:
                if not self.interrupted:
                    raise RuntimeError(
                        f"Simpy interrupt occured at {self.data.ID} although process is not interrupted"
                    )
        while self.done_in:
            try:
                if self.interrupted:
                    self.state_info.log_start_interrupt_state(
                        self.env.now, StateTypeEnum.transport
                    )
                    self.update_done_in()
                    yield events.AllOf(self.env, [self.active, self.resource.active])
                    self.interrupted = False
                    self.state_info.log_end_interrupt_state(
                        self.env.now,
                        self.env.now + self.done_in,
                        StateTypeEnum.transport,
                    )
                self.start = self.env.now
                self.state_info.log_start_state(
                    self.start, self.start + self.done_in, StateTypeEnum.transport
                )
                yield self.env.timeout(self.done_in)
                self.done_in = 0  # Set to 0 to exit while loop.
            except exceptions.Interrupt:
                if not self.interrupted:
                    raise RuntimeError(
                        f"Simpy interrupt occured at {self.data.ID} although process is not interrupted"
                    )
        self.state_info.log_end_state(self.env.now, StateTypeEnum.transport)

    def update_done_in(self):
        self.done_in -= self.env.now - self.start  # How much time left?
        if self.done_in < 0:
            self.done_in = 0

    def interrupt_process(self):
        if self.process and self.process.is_alive and not self.interrupted:
            self.interrupted = True
            self.process.interrupt()


class BreakDownState(State):
    """
    Represents a breakdown state of a resource in the simulation. A breakdown state has a process that simulates the breakdown of a resource. All other running production, transport or setup states get interrupted.

    Args:
        data (BreakDownStateData): The data of the state.
        time_model (time_model.TimeModel): The time model of the state.
        env (sim.Environment): The simulation environment.
        active (events.Event, optional): Event that indidcates if the state is active. Defaults to None.
        resource (resources.Resource, optional): The resource the state belongs to. Defaults to None.
        process (Optional[events.Process], optional): The process of the state. Defaults to None.
        state_info (StateInfo, optional): The state information of the state. Defaults to None.
        repair_time_model (time_model.TimeModel, optional): The time model of the repair time. Defaults to None.
    """

    def __init__(
        self,
        data: BreakDownStateData,
        time_model: time_model.TimeModel,
        env: sim.Environment,
        repair_time_model: Optional[time_model.TimeModel] = None,
    ):
        super().__init__(data, time_model, env)
        self.repair_time_model = repair_time_model
        self.active_breakdown = False
        self.active = events.Event(self.env)

    def process_state(self) -> Generator:
        while True:
            yield self.env.process(self.wait_for_breakdown())
            self.active_breakdown = True
            self.resource.interrupt_states()
            repair_time = self.repair_time_model.get_next_time()
            self.state_info.log_start_state(
                self.env.now, self.env.now + repair_time, StateTypeEnum.breakdown
            )
            yield self.env.timeout(repair_time)
            self.active_breakdown = False
            self.resource.activate()
            self.state_info.log_end_state(self.env.now, StateTypeEnum.breakdown)

    def wait_for_breakdown(self):
        yield self.env.timeout(self.time_model.get_next_time())

    def interrupt_process(self):
        pass


class NonScheduledState(State):
    """
    Represents a non-scheduled state of a resource in the simulation. This state models shift availability
    by alternating between scheduled (when the resource is available) and non-scheduled (when the resource
    is unavailable/shift is off) time intervals. During non-scheduled intervals, all operations are blocked
    similar to a breakdown state.

    Args:
        data (NonScheduledStateData): The data of the state.
        time_model (time_model.TimeModel): The time model for scheduled intervals (when resource is available).
        env (sim.Environment): The simulation environment.
        active (events.Event, optional): Event that indicates if the state is active. Defaults to None.
        resource (resources.Resource, optional): The resource the state belongs to. Defaults to None.
        process (Optional[events.Process], optional): The process of the state. Defaults to None.
        state_info (StateInfo, optional): The state information of the state. Defaults to None.
        non_scheduled_time_model (time_model.TimeModel, optional): The time model for non-scheduled intervals (when resource is unavailable). Defaults to None.
    """

    def __init__(
        self,
        data: NonScheduledStateData,
        time_model: time_model.TimeModel,
        env: sim.Environment,
        non_scheduled_time_model: Optional[time_model.TimeModel] = None,
    ):
        super().__init__(data, time_model, env)
        self.non_scheduled_time_model = non_scheduled_time_model
        self.active_non_scheduled = False
        self.active = events.Event(self.env)

    def process_state(self) -> Generator:
        while True:
            # Scheduled period: resource is available (wait for scheduled time)
            scheduled_time = self.time_model.get_next_time()
            yield self.env.timeout(scheduled_time)
            
            # Non-scheduled period: resource is unavailable (blocks all operations)
            self.active_non_scheduled = True
            self.resource.interrupt_states()
            non_scheduled_time = self.non_scheduled_time_model.get_next_time()
            self.state_info.log_start_state(
                self.env.now, self.env.now + non_scheduled_time, StateTypeEnum.non_scheduled
            )
            yield self.env.timeout(non_scheduled_time)
            self.active_non_scheduled = False
            self.resource.activate()
            self.state_info.log_end_state(self.env.now, StateTypeEnum.non_scheduled)

    def interrupt_process(self):
        pass


class ProcessBreakDownState(State):
    """
    Represents a process breakdown state of a resource in the simulation. A process breakdown state has a process that simulates the breakdown of a process of a resource. Only production states of this type of process get interrupted. Also all setup states get interrupted.

    Args:
        data (ProcessBreakDownStateData): The data of the state.
        time_model (time_model.TimeModel): The time model of the state.
        env (sim.Environment): The simulation environment.
        active (events.Event, optional): Event that indidcates if the state is active. Defaults to None.
        resource (resources.Resource, optional): The resource the state belongs to. Defaults to None.
        process (Optional[events.Process], optional): The process of the state. Defaults to None.
        state_info (StateInfo, optional): The state information of the state. Defaults to None.
        production_states (List[State], optional): The production states of the process. Defaults to None.
        repair_time_model (time_model.TimeModel, optional): The time model of the repair time. Defaults to None.
    """

    def __init__(
        self,
        data: ProcessBreakDownStateData,
        time_model: time_model.TimeModel,
        env: sim.Environment,
        repair_time_model: Optional[time_model.TimeModel] = None,
    ):
        super().__init__(data, time_model, env)
        self.repair_time_model = repair_time_model
        self.production_states = []

    def set_production_states(self, production_states: List[ProductionState]):
        if any(
            [
                production_state.data.ID != self.data.process_id
                for production_state in production_states
            ]
        ):
            raise ValueError(
                f"Production states {production_states} do not match process id {self.data.process_id}"
            )
        self.production_states = production_states

    def process_state(self) -> Generator:
        while True:
            yield self.env.process(self.wait_for_breakdown())
            yield events.AllOf(
                self.env, [state.active for state in self.production_states]
            )
            for state in self.production_states + self.resource.setup_states:
                state.deactivate()
            for state in self.production_states + self.resource.setup_states:
                if state.process and state.process.is_alive and not state.interrupted:
                    state.interrupt_process()
            repair_time = self.repair_time_model.get_next_time()
            self.state_info.log_start_state(
                self.env.now,
                self.env.now + repair_time,
                StateTypeEnum.process_breakdown,
            )
            yield self.env.timeout(repair_time)
            self.state_info.log_end_state(self.env.now, StateTypeEnum.process_breakdown)
            for state in self.production_states + self.resource.setup_states:
                state.activate()

    def wait_for_breakdown(self):
        yield self.env.timeout(self.time_model.get_next_time())

    def interrupt_process(self):
        pass


class SetupState(State):
    """
    Represents a setup state of a resource in the simulation. A setup state has a process that simulates the setup of a resource. This changes the current setup of the resource and allows it processing of other types of processes with their associated production or transport states.

    Args:
        data (SetupStateData): The data of the state.
        time_model (time_model.TimeModel): The time model of the state.
        env (sim.Environment): The simulation environment.
        active (events.Event, optional): Event that indidcates if the state is active. Defaults to None.
        resource (resources.Resource, optional): The resource the state belongs to. Defaults to None.
        process (Optional[events.Process], optional): The process of the state. Defaults to None.
        state_info (StateInfo, optional): The state information of the state. Defaults to None.
        start (float, optional): The start time of the state. Defaults to 0.0.
        done_in (float, optional): The time the state is done in. Defaults to 0.0.

    Attributes:
        interrupt_processed (events.Event): Event that indicates if the state is interrupted. Defaults to None.
    """

    def __init__(
        self,
        data: SetupStateData,
        time_model: time_model.TimeModel,
        env: sim.Environment,
    ):
        super().__init__(data, time_model, env)
        self.start = 0.0
        self.done_in = 0.0
        self.interrupted = False

    def process_state(self) -> Generator:
        self.done_in = self.time_model.get_next_time()
        while True:
            try:
                if self.interrupted:
                    yield events.AllOf(self.env, [self.active, self.resource.active])
                    self.interrupted = False
                yield events.AllOf(self.env, [self.resource.active, self.active])
                running_processes = [
                    state.process
                    for state in self.resource.production_states
                    if (state.process and state.process.is_alive)
                ]
                yield events.AllOf(self.env, running_processes)
                break
            except exceptions.Interrupt:
                if not self.interrupted:
                    raise RuntimeError(
                        f"Simpy interrupt occured at {self.data.ID} although process is not interrupted"
                    )
        while self.done_in:
            try:
                if self.interrupted:
                    self.state_info.log_start_interrupt_state(
                        self.env.now, StateTypeEnum.setup
                    )
                    self.update_done_in()
                    yield events.AllOf(self.env, [self.active, self.resource.active])
                    self.interrupted = False
                    self.state_info.log_end_interrupt_state(
                        self.env.now, self.env.now + self.done_in, StateTypeEnum.setup
                    )
                self.start = self.env.now
                self.state_info.log_start_state(
                    self.start, self.start + self.done_in, StateTypeEnum.setup
                )
                yield self.env.timeout(self.done_in)
                self.done_in = 0
            except exceptions.Interrupt:
                if not self.interrupted:
                    raise RuntimeError(
                        f"Simpy interrupt occured at {self.data.ID} although process is not interrupted"
                    )
        self.state_info.log_end_state(self.env.now, StateTypeEnum.setup)

    def update_done_in(self):
        self.done_in -= self.env.now - self.start  # How much time left?
        if self.done_in < 0:
            self.done_in = 0

    def interrupt_process(self) -> Generator:
        if self.process and self.process.is_alive and not self.interrupted:
            self.interrupted = True
            self.process.interrupt()


MINIMUM_BATTERY_LEVEL = 0.1


class ChargingState(State):
    """
    Represents a charging state of a resource in the simulation. A charging state has a process that simulates the charging of a resource. This changes the current charge of the resource and allows it processing of other types of processes with their associated production or transport

    Args:
        data (ChargingStateData): The data of the state.
        time_model (time_model.TimeModel): The time model of the state.
        battery_time_model (time_model.TimeModel): The time model of the battery.
        env (sim.Environment): The simulation environment.
        active (events.Event, optional): Event that indidcates if the state is active. Defaults to None.
        resource (resources.Resource, optional): The resource the state belongs to. Defaults to None.
        process (Optional[events.Process], optional): The process of the state. Defaults to None.
        state_info (StateInfo, optional): The state information of the state. Defaults to None.
        start (float, optional): The start time of the state. Defaults to 0.0.
        done_in (float, optional): The ramaining time for the state to finish. Defaults to 0.0.
        interrupted (bool, optional): Indicates if the state is interrupted. Defaults to False.
    """

    def __init__(
        self,
        data: ChargingStateData,
        time_model: time_model.TimeModel,
        battery_time_model: time_model.TimeModel,
        env: sim.Environment,
    ):
        super().__init__(data, time_model, env)
        self.battery_time_model = battery_time_model
        self.battery_usage_time_since_charging = 0.0

        self.start = 0.0
        self.done_in = 0.0
        self.interrupted = False

    def requires_charging(self) -> bool:
        return (
            self.battery_usage_time_since_charging
            >= (1 - MINIMUM_BATTERY_LEVEL) * self.battery_time_model.get_next_time()
        )

    def add_battery_usage_time(self, time: float):
        self.battery_usage_time_since_charging += time

    def process_state(self) -> Generator:
        self.done_in = self.time_model.get_next_time()
        while True:
            try:
                if self.interrupted:
                    yield events.AllOf(self.env, [self.active, self.resource.active])
                    self.interrupted = False
                yield events.AllOf(self.env, [self.resource.active, self.active])
                break
            except exceptions.Interrupt:
                if not self.interrupted:
                    raise RuntimeError(
                        f"Simpy interrupt occured at {self.data.ID} although process is not interrupted"
                    )
        while self.done_in:
            try:
                if self.interrupted:
                    self.state_info.log_start_interrupt_state(
                        self.env.now, StateTypeEnum.production
                    )
                    self.update_done_in()
                    yield events.AllOf(self.env, [self.active, self.resource.active])
                    self.interrupted = False
                    self.state_info.log_end_interrupt_state(
                        self.env.now,
                        self.env.now + self.done_in,
                        StateTypeEnum.charging,
                    )
                self.start = self.env.now
                self.state_info.log_start_state(
                    self.start, self.start + self.done_in, StateTypeEnum.charging
                )
                yield self.env.timeout(self.done_in)
                self.done_in = 0  # Set to 0 to exit while loop.

            except exceptions.Interrupt:
                if not self.interrupted:
                    raise RuntimeError(
                        f"Simpy interrupt occured at {self.data.ID} although process is not interrupted"
                    )
        self.state_info.log_end_state(self.env.now, StateTypeEnum.charging)
        self.battery_usage_time_since_charging = 0

    def update_done_in(self):
        if self.start == 0:
            return
        self.done_in -= self.env.now - self.start  # How much time left?
        if self.done_in < 0:
            self.done_in = 0

    def interrupt_process(self):
        if self.process and self.process.is_alive and not self.interrupted:
            self.interrupted = True
            self.process.interrupt()


STATE_UNION = Union[
    ChargingState,
    BreakDownState,
    ProductionState,
    TransportState,
    SetupState,
    ProcessBreakDownState,
]
"""
Union Type of all states.
"""

if TYPE_CHECKING:
    from prodsys.simulation import product



from __future__ import annotations

from typing import Dict, List, Optional, Tuple

import heapq
from pathfinding.core.graph import GraphNode


class DijkstraAllPaths:
    """
    Efficient Dijkstra implementation that explores all nodes from an origin
    and caches all shortest paths. This allows O(1) path retrieval after the
    initial exploration.
    """
    
    def __init__(self):
        """
        Initializes the Dijkstra path finder with empty caches.
        """
        # Cache: (process_id, origin_id) -> Dict[target_id, List[GraphNode]]
        self._path_cache: Dict[Tuple[str, str], Dict[str, List[GraphNode]]] = {}
        # Cache: (process_id, origin_id) -> adjacency list representation
        self._graph_cache: Dict[Tuple[str, str], Dict[str, List[Tuple[str, float]]]] = {}
    
    def get_all_paths_from_origin(
        self,
        origin: GraphNode,
        edges: List[Tuple[GraphNode, GraphNode, float]],
        process_id: str,
    ) -> Dict[str, List[GraphNode]]:
        """
        Explores all nodes from the origin using Dijkstra's algorithm and returns
        all shortest paths. Results are cached for efficient subsequent lookups.
        
        Args:
            origin (GraphNode): The origin node to start exploration from.
            edges (List[Tuple[GraphNode, GraphNode, float]]): List of edges as
                (from_node, to_node, cost) tuples.
            process_id (str): The process ID for caching purposes.
            
        Returns:
            Dict[str, List[GraphNode]]: Dictionary mapping target node IDs to
                their shortest paths from origin.
        """
        origin_id = origin.node_id
        cache_key = (process_id, origin_id)
        
        # Check if we already have paths cached for this origin
        if cache_key in self._path_cache:
            return self._path_cache[cache_key]
        
        # Build adjacency list from edges and collect all nodes
        adjacency_list, node_map = self._build_adjacency_list(edges)
        # Ensure origin is in node_map
        node_map[origin_id] = origin
        
        # Run Dijkstra's algorithm to find all shortest paths
        paths = self._dijkstra_explore_all(origin, adjacency_list, node_map)
        
        # Cache the results
        self._path_cache[cache_key] = paths
        self._graph_cache[cache_key] = adjacency_list
        
        return paths
    
    def _build_adjacency_list(
        self, edges: List[Tuple[GraphNode, GraphNode, float]]
    ) -> Tuple[Dict[str, List[Tuple[str, float]]], Dict[str, GraphNode]]:
        """
        Builds an adjacency list representation from edges and collects all nodes.
        
        Args:
            edges (List[Tuple[GraphNode, GraphNode, float]]): List of edges.
            
        Returns:
            Tuple[Dict[str, List[Tuple[str, float]]], Dict[str, GraphNode]]: 
                A tuple containing:
                - Adjacency list mapping node_id to list of (neighbor_id, cost) tuples.
                - Dictionary mapping node_id to GraphNode objects.
        """
        adjacency_list: Dict[str, List[Tuple[str, float]]] = {}
        node_map: Dict[str, GraphNode] = {}
        
        for from_node, to_node, cost in edges:
            from_id = from_node.node_id
            to_id = to_node.node_id
            
            # Store nodes in node_map
            node_map[from_id] = from_node
            node_map[to_id] = to_node
            
            # Add forward edge
            if from_id not in adjacency_list:
                adjacency_list[from_id] = []
            adjacency_list[from_id].append((to_id, cost))
            
            # Add backward edge (bi-directional graph)
            if to_id not in adjacency_list:
                adjacency_list[to_id] = []
            adjacency_list[to_id].append((from_id, cost))
        
        return adjacency_list, node_map
    
    def _dijkstra_explore_all(
        self,
        origin: GraphNode,
        adjacency_list: Dict[str, List[Tuple[str, float]]],
        node_map: Dict[str, GraphNode],
    ) -> Dict[str, List[GraphNode]]:
        """
        Runs Dijkstra's algorithm to explore all reachable nodes from origin
        and returns all shortest paths.
        
        Args:
            origin (GraphNode): The origin node.
            adjacency_list (Dict[str, List[Tuple[str, float]]]): Adjacency list.
            node_map (Dict[str, GraphNode]): Mapping from node IDs to GraphNodes.
            
        Returns:
            Dict[str, List[GraphNode]]: Dictionary mapping target node IDs to
                their shortest paths from origin.
        """
        origin_id = origin.node_id
        
        # Initialize distances and previous nodes
        distances: Dict[str, float] = {origin_id: 0.0}
        previous: Dict[str, Optional[str]] = {origin_id: None}
        
        # Priority queue: (distance, node_id)
        pq: List[Tuple[float, str]] = [(0.0, origin_id)]
        visited: set[str] = set()
        
        # Ensure origin is in node_map
        node_map[origin_id] = origin
        
        while pq:
            current_dist, current_id = heapq.heappop(pq)
            
            if current_id in visited:
                continue
            
            visited.add(current_id)
            
            # Explore neighbors
            if current_id in adjacency_list:
                for neighbor_id, edge_cost in adjacency_list[current_id]:
                    if neighbor_id in visited:
                        continue
                    
                    new_dist = current_dist + edge_cost
                    
                    # Update if we found a shorter path
                    if neighbor_id not in distances or new_dist < distances[neighbor_id]:
                        distances[neighbor_id] = new_dist
                        previous[neighbor_id] = current_id
                        heapq.heappush(pq, (new_dist, neighbor_id))
                        
                        # Create GraphNode if not exists (shouldn't happen if edges are complete)
                        if neighbor_id not in node_map:
                            node_map[neighbor_id] = GraphNode(node_id=neighbor_id)
        
        # Reconstruct all paths
        paths: Dict[str, List[GraphNode]] = {}
        
        for target_id in visited:
            path = self._reconstruct_path(origin_id, target_id, previous, node_map)
            if path:
                paths[target_id] = path
        
        return paths
    
    def _reconstruct_path(
        self,
        origin_id: str,
        target_id: str,
        previous: Dict[str, Optional[str]],
        node_map: Dict[str, GraphNode],
    ) -> List[GraphNode]:
        """
        Reconstructs the shortest path from origin to target.
        
        Args:
            origin_id (str): Origin node ID.
            target_id (str): Target node ID.
            previous (Dict[str, Optional[str]]): Dictionary mapping node IDs to
                their previous node IDs in the shortest path.
            node_map (Dict[str, GraphNode]): Mapping from node IDs to GraphNodes.
            
        Returns:
            List[GraphNode]: The shortest path as a list of GraphNodes.
        """
        if target_id not in previous:
            return []
        
        path = []
        current_id = target_id
        
        # Reconstruct path backwards
        while current_id is not None:
            if current_id not in node_map:
                # Create GraphNode if missing
                node_map[current_id] = GraphNode(node_id=current_id)
            path.append(node_map[current_id])
            current_id = previous.get(current_id)
        
        # Reverse to get path from origin to target
        path.reverse()
        return path
    
    def get_path(
        self,
        origin: GraphNode,
        target: GraphNode,
        edges: List[Tuple[GraphNode, GraphNode, float]],
        process_id: str,
    ) -> Optional[List[GraphNode]]:
        """
        Gets the shortest path from origin to target, using cached results
        if available.
        
        Args:
            origin (GraphNode): Origin node.
            target (GraphNode): Target node.
            edges (List[Tuple[GraphNode, GraphNode, float]]): List of edges.
            process_id (str): Process ID for caching.
            
        Returns:
            Optional[List[GraphNode]]: The shortest path, or None if no path exists.
        """
        # Early termination if origin and target are the same
        if origin.node_id == target.node_id:
            return [origin]
        
        # Get all paths from origin (will use cache if available)
        all_paths = self.get_all_paths_from_origin(origin, edges, process_id)
        
        # Return the path to target if it exists
        return all_paths.get(target.node_id)



import pytest
from prodsys.models.production_system_data import ProductionSystemData
import prodsys.express as psx
from prodsys import runner
from prodsys.simulation import sim


# Define the path to your test configuration file. This should be similar to example_simulation.py.
@pytest.fixture
def simulation_adapter() -> ProductionSystemData:
    time_model_agv = psx.DistanceTimeModel(
        speed=90, reaction_time=0.2, ID="time_model_x"
    )

    time_model_machine1 = psx.FunctionTimeModel(
        distribution_function="constant", location=3, ID="time_model_ap23"
    )
    time_model_machine2 = psx.FunctionTimeModel(
        distribution_function="constant", location=3, ID="time_model_ap23"
    )
    time_model_machine3 = psx.FunctionTimeModel(
        distribution_function="constant", location=6, ID="time_model_ap23"
    )
    time_model_machine4 = psx.FunctionTimeModel(
        distribution_function="constant", location=6, ID="time_model_ap23"
    )
    time_model_machine5 = psx.FunctionTimeModel(
        distribution_function="constant", location=6, ID="time_model_ap23"
    )
    time_model_machine6 = psx.FunctionTimeModel(
        distribution_function="constant", location=3, ID="time_model_ap23"
    )
    timer_model_interarrival_time = psx.FunctionTimeModel(
        distribution_function="constant", location=6, ID="time_model_source01"
    )
    timer_model_interarrival_time2 = psx.FunctionTimeModel(
        distribution_function="constant", location=6, ID="time_model_source02"
    )

    node1 = psx.Node(location=[0, 20], ID="node1")
    node2 = psx.Node(location=[50, 20], ID="node2")
    node3 = psx.Node(location=[100, 20], ID="node3")
    node4 = psx.Node(location=[100, 80], ID="node4")
    node5 = psx.Node(location=[50, 80], ID="node5")
    node6 = psx.Node(location=[0, 80], ID="node6")

    rcp01 = psx.RequiredCapabilityProcess(
        time_model=time_model_agv, capability="euro_palette_transport", ID="rtp01"
    )
    productionprocess01 = psx.ProductionProcess(
        time_model=time_model_machine1, ID="pp01"
    )
    productionprocess02 = psx.ProductionProcess(
        time_model=time_model_machine2, ID="pp02"
    )
    productionprocess03 = psx.ProductionProcess(
        time_model=time_model_machine3, ID="pp03"
    )
    productionprocess04 = psx.ProductionProcess(
        time_model=time_model_machine4, ID="pp04"
    )
    productionprocess05 = psx.ProductionProcess(
        time_model=time_model_machine5, ID="pp05"
    )
    productionprocess06 = psx.ProductionProcess(
        time_model=time_model_machine6, ID="pp06"
    )

    machine01 = psx.Resource(
        ID="resource01",
        processes=[productionprocess01],
        location=[0, 0],
    )
    machine02 = psx.Resource(
        ID="resource02",
        processes=[productionprocess02],
        location=[50, 0],
    )
    machine03 = psx.Resource(
        ID="resource03",
        processes=[productionprocess03],
        location=[100, 0],
    )
    machine04 = psx.Resource(
        ID="resource04",
        processes=[productionprocess04],
        location=[100, 100],
    )
    machine05 = psx.Resource(
        ID="resource05",
        processes=[productionprocess05],
        location=[50, 100],
    )
    machine06 = psx.Resource(
        ID="resource06",
        processes=[productionprocess06],
        location=[0, 100],
    )

    product01 = psx.Product(
        process=[
            productionprocess01,
            productionprocess02,
            productionprocess03,
            productionprocess04,
            productionprocess05,
            productionprocess06,
        ],
        transport_process=rcp01,
        ID="product01",
    )

    product02 = psx.Product(
        process=[
            productionprocess01,
            productionprocess02,
            productionprocess03,
            productionprocess04,
            productionprocess06,
        ],
        transport_process=rcp01,
        ID="product02",
    )

    source01 = psx.Source(
        product=product01,
        ID="source01",
        time_model=timer_model_interarrival_time,
        location=[-10, 0],
    )

    source02 = psx.Source(
        product=product02,
        ID="source02",
        time_model=timer_model_interarrival_time2,
        location=[-10, 0],
    )
    sink01 = psx.Sink(product=product01, ID="sink01", location=[-10, 100])
    sink02 = psx.Sink(product=product02, ID="sink02", location=[-10, 100])

    ltp01_links = [
        [source01, machine01],
        [source02, machine01],
        [machine01, node1],
        [node1, node2],
        [node2, machine02],
        [node2, node3],
        [node3, machine03],
    ]

    ltp01 = psx.LinkTransportProcess(
        time_model=time_model_agv,
        capability="euro_palette_transport",
        ID="ltp01",
        links=ltp01_links,
    )

    ltp02_links = [
        [node3, machine03],
        [node3, node4],
        [node4, machine04],
        [node4, node5],
        [node5, machine05],
        [node5, node6],
        [node6, machine06],
        [machine06, sink01],
        [machine06, sink02],
    ]

    ltp02 = psx.LinkTransportProcess(
        time_model=time_model_agv,
        capability="euro_palette_transport",
        ID="ltp02",
        links=ltp02_links,
    )

    agv01 = psx.Resource(
        location=[50, 20],
        ID="agv01",
        processes=[ltp01],
    )

    agv02 = psx.Resource(
        location=[50, 20],
        ID="agv02",
        processes=[ltp01],
    )

    agv03 = psx.Resource(
        location=[50, 20],
        ID="agv03",
        processes=[ltp01],
    )

    agv04 = psx.Resource(
        location=[50, 80],
        ID="agv04",
        processes=[ltp02],
    )

    agv05 = psx.Resource(
        location=[50, 80],
        ID="agv05",
        processes=[ltp02],
    )

    agv06 = psx.Resource(
        location=[50, 80],
        ID="agv06",
        processes=[ltp02],
    )

    productionsystem = psx.ProductionSystem(
        resources=[
            agv01,
            agv02,
            agv03,
            agv04,
            agv05,
            agv06,
            machine01,
            machine02,
            machine03,
            machine04,
            machine05,
            machine06,
        ],
        sources=[source01, source02],
        sinks=[sink01, sink02],
        ID="productionsystem01",
    )

    adapter = productionsystem.to_model()
    return adapter


def test_validate_simulation(simulation_adapter: ProductionSystemData):
    simulation_adapter.validate_configuration()


def test_initialize_simulation(simulation_adapter: ProductionSystemData):
    runner_instance = runner.Runner(production_system_data=simulation_adapter)
    runner_instance.initialize_simulation()


def test_run_simulation(simulation_adapter: ProductionSystemData):
    runner_instance = runner.Runner(production_system_data=simulation_adapter)
    runner_instance.initialize_simulation()
    runner_instance.run(480)
    assert runner_instance.env.now == 480
    post_processor = runner_instance.get_post_processor()
    for counter, kpi in enumerate(post_processor.throughput_and_output_KPIs):
        if kpi.name == "output":
            assert kpi.value > 1
    assert counter == 2 * 2 - 1

    productive_counter = 0
    for counter, kpi in enumerate(post_processor.machine_state_KPIS):
        if kpi.name == "productive_time":
            productive_counter += 1
            assert kpi.value > 1 and kpi.value < 99
    assert productive_counter == 12

    for counter, kpi in enumerate(post_processor.WIP_KPIs):
        if kpi.name == "WIP":
            assert kpi.value > 0.01 and kpi.value < 500

    assert counter == 2 + 1 - 1



from typing import TYPE_CHECKING, List

from prodsys.models.node_data import NodeData
from prodsys.simulation import sim

class Node:
    def __init__(self, data: NodeData, env: sim.Environment):
        """
        Initialize the Node with the given data.

        Args:
            data (NodeData): The data for the node.
            env (sim.Environment): The environment of the node.
        """
        self.env = env
        self.data = data

    def get_location(self) -> List[float]:
        """
        Returns the location of the resource.

        Returns:
            List[float]: The location of the resource. Has to have length 2.
        """
        return self.data.location



from __future__ import annotations

from typing import Optional, Protocol, TYPE_CHECKING, runtime_checkable

import logging



from prodsys.simulation.entities import primitive
from prodsys.simulation import state
if TYPE_CHECKING:
    from prodsys.simulation import resources

    from prodsys.models import dependency_data
    from prodsys.simulation import (
        process,
        sim,
        state,
        node,
    )

logger = logging.getLogger(__name__)

class DependedEntity(Protocol):
    """
    Protocol that defines the interface for a depended entity. This is used to define the type of the depended entity in the Dependency class.
    """

    def bind(self, dependant: process.Process | resources.Resource | primitive.Primitive, dependency: Dependency) -> None:
        """
        Binds the depended entity to the dependency.
        """
        pass
        # When binding the items, they should be transport to the location of destination
        # they should go in a status so that they are not used elsewhere, until release is called.

    def release(self):
        """
        Releases the product object.
        """
        pass
        # release should make DependedEntity available for other processes
        # called by the dependant


class DependencyInfo:
    """
    Class that represents information of the current state of a resource.

    Args:
        resource_ID (str): ID of the resource that the product is currently at.
        state_ID (str): ID of the state that the product is currently at.
        event_time (float): Time of the event.
        activity (state.StateEnum): Activity of the product.
        product_ID (str): ID of the product.
        state_type (state.StateTypeEnum): Type of the state.
    """

    def __init__(self, resource_id: Optional[str]=None, primitive_id: Optional[str]=None):
        """
        Initializes the DependencyInfo class.
        """
        self.resource_ID: str = resource_id
        self.state_ID: str = "Dependency"
        self.event_time: float = None
        self.activity: state.StateEnum = None
        self.primitive_ID: str = primitive_id
        self.state_type: state.StateTypeEnum = state.StateTypeEnum.dependency
        # self.state_type: state.StateTypeEnum = state.StateTypeEnum.production
        self.requesting_item_ID: str = None
        self.dependency_ID: str = None

    def log_start_dependency(
        self,
        event_time: float,
        requesting_item_id: str,
        dependency_id: str,
    ) -> None:
        """
        Logs the start of a dependency.

        Args:
            event_time (float): Time of the event.
            activity (state.StateEnum): Activity of the product.
            requesting_item (str): ID of the product that is requesting the dependency.
        """
        self.event_time = event_time
        self.activity = state.StateEnum.start_state
        self.requesting_item_ID = requesting_item_id
        self.dependency_ID = dependency_id

    def log_end_dependency(
        self,
        event_time: float,
        requesting_item_id: str,
        dependency_id: str,
    ) -> None:
        """
        Logs the end of a dependency.

        Args:
            event_time (float): Time of the event.
            activity (state.StateEnum): Activity of the product.
            requesting_item (str): ID of the product that is requesting the dependency.
        """
        self.event_time = event_time
        self.activity = state.StateEnum.end_state
        self.requesting_item_ID = requesting_item_id
        self.dependency_ID = dependency_id


class Dependency:
    """
    Class that represents a dependency in the discrete event simulation. For easier instantion of the class, use the PrimitiveFactory at prodsys.factories.primitive.
    """

    def __init__(
        self,
        env: sim.Environment,
        data: dependency_data.DependencyData,
        required_process: Optional[process.Process],
        required_entity: Optional[primitive.Primitive],
        required_resource: Optional[resources.Resource],
        interaction_node: Optional[node.Node],
    ):
        """
        Initializes the Dependency class.

        Args:
            env (sim.Environment): prodsys simulation environment.
            data (dependency_data.DependencyData): Dependency data of the product.
            required_process (process.Process): Required process of the product.
            required_primitive (primitive.Primitive): Required primitive of the product.
            required_resource (resources.Resource): Required resource of the product.
            interaction_node (node.Node): Interaction node of the product.
        """
        self.env = env
        self.data = data
        self.required_process = required_process
        self.required_entity = required_entity
        self.required_resource = required_resource
        self.interaction_node = interaction_node





"""
Tests for state_data module.
"""

import pytest
from prodsys.models.state_data import (
    StateData,
    BreakDownStateData,
    ProcessBreakDownStateData,
    ProductionStateData,
    TransportStateData,
    SetupStateData,
    ChargingStateData,
    StateTypeEnum,
)
from prodsys.models.processes_data import ProductionProcessData, ProcessTypeEnum
from prodsys.models.time_model_data import FunctionTimeModelData
from prodsys.models.production_system_data import ProductionSystemData
from prodsys.util.statistical_functions import FunctionTimeModelEnum


@pytest.fixture
def sample_time_model():
    """Create a sample time model for testing."""
    return FunctionTimeModelData(
        ID="tm1",
        description="Test time model",
        distribution_function=FunctionTimeModelEnum.Exponential,
        location=540.0,
        scale=0.0,
    )


@pytest.fixture
def sample_repair_time_model():
    """Create a sample repair time model for testing."""
    return FunctionTimeModelData(
        ID="tm_repair",
        description="Repair time model",
        distribution_function=FunctionTimeModelEnum.Exponential,
        location=60.0,
        scale=0.0,
    )


@pytest.fixture
def sample_process_time_model():
    """Create a sample process time model for testing."""
    return FunctionTimeModelData(
        ID="tm_process",
        description="Process time model",
        distribution_function=FunctionTimeModelEnum.Normal,
        location=20.0,
        scale=5.0,
    )


@pytest.fixture
def sample_process(sample_process_time_model):
    """Create a sample process for testing."""
    return ProductionProcessData(
        ID="P1",
        description="Process 1",
        time_model_id="tm_process",
        type=ProcessTypeEnum.ProductionProcesses,
    )


@pytest.fixture
def sample_production_system(sample_time_model, sample_repair_time_model, sample_process_time_model, sample_process):
    """Create a minimal production system for testing."""
    return ProductionSystemData(
        time_model_data=[sample_time_model, sample_repair_time_model, sample_process_time_model],
        process_data=[sample_process],
        resource_data=[],
        product_data=[],
        source_data=[],
        sink_data=[],
        state_data=[],
        port_data=[],
        primitive_data=[],
        dependency_data=[],
        node_data=[],
        scenario_data=None,
    )


class TestBreakDownStateData:
    """Tests for BreakDownStateData."""

    def test_creation_with_required_fields(self):
        """Test creating BreakDownStateData with required fields."""
        state = BreakDownStateData(
            ID="Breakdownstate_1",
            description="Breakdown state 1",
            time_model_id="tm1",
            type=StateTypeEnum.BreakDownState,
            repair_time_model_id="tm_repair",
        )
        assert state.ID == "Breakdownstate_1"
        assert state.description == "Breakdown state 1"
        assert state.time_model_id == "tm1"
        assert state.type == StateTypeEnum.BreakDownState
        assert state.repair_time_model_id == "tm_repair"

    def test_hash(self, sample_production_system, sample_time_model, sample_repair_time_model):
        """Test hash method."""
        state1 = BreakDownStateData(
            ID="Breakdownstate_1",
            description="Breakdown state 1",
            time_model_id="tm1",
            type=StateTypeEnum.BreakDownState,
            repair_time_model_id="tm_repair",
        )
        state2 = BreakDownStateData(
            ID="Breakdownstate_2",
            description="Breakdown state 2",
            time_model_id="tm1",
            type=StateTypeEnum.BreakDownState,
            repair_time_model_id="tm_repair",
        )
        
        hash1 = state1.hash(sample_production_system)
        hash2 = state2.hash(sample_production_system)
        
        # Same parameters should produce same hash
        assert hash1 == hash2


class TestProcessBreakDownStateData:
    """Tests for ProcessBreakDownStateData."""

    def test_creation_with_required_fields(self):
        """Test creating ProcessBreakDownStateData with required fields."""
        state = ProcessBreakDownStateData(
            ID="ProcessBreakDownState_1",
            description="Process Breakdown state 1",
            time_model_id="tm1",
            type=StateTypeEnum.ProcessBreakDownState,
            repair_time_model_id="tm_repair",
            process_id="P1",
        )
        assert state.ID == "ProcessBreakDownState_1"
        assert state.time_model_id == "tm1"
        assert state.type == StateTypeEnum.ProcessBreakDownState
        assert state.repair_time_model_id == "tm_repair"
        assert state.process_id == "P1"

    def test_hash(self, sample_production_system, sample_time_model, sample_repair_time_model, sample_process):
        """Test hash method."""
        state1 = ProcessBreakDownStateData(
            ID="ProcessBreakDownState_1",
            description="Process Breakdown state 1",
            time_model_id="tm1",
            type=StateTypeEnum.ProcessBreakDownState,
            repair_time_model_id="tm_repair",
            process_id="P1",
        )
        state2 = ProcessBreakDownStateData(
            ID="ProcessBreakDownState_2",
            description="Process Breakdown state 2",
            time_model_id="tm1",
            type=StateTypeEnum.ProcessBreakDownState,
            repair_time_model_id="tm_repair",
            process_id="P1",
        )
        
        hash1 = state1.hash(sample_production_system)
        hash2 = state2.hash(sample_production_system)
        
        # Same parameters should produce same hash
        assert hash1 == hash2


class TestProductionStateData:
    """Tests for ProductionStateData."""

    def test_creation_with_required_fields(self):
        """Test creating ProductionStateData with required fields."""
        state = ProductionStateData(
            ID="ProductionState_1",
            description="Production state 1",
            time_model_id="tm1",
            type=StateTypeEnum.ProductionState,
        )
        assert state.ID == "ProductionState_1"
        assert state.description == "Production state 1"
        assert state.time_model_id == "tm1"
        assert state.type == StateTypeEnum.ProductionState


class TestTransportStateData:
    """Tests for TransportStateData."""

    def test_creation_with_required_fields(self):
        """Test creating TransportStateData with required fields."""
        state = TransportStateData(
            ID="TransportState_1",
            description="Transport state 1",
            time_model_id="tm1",
            type=StateTypeEnum.TransportState,
        )
        assert state.ID == "TransportState_1"
        assert state.time_model_id == "tm1"
        assert state.type == StateTypeEnum.TransportState

    def test_default_loading_unloading_time_models(self):
        """Test that loading/unloading time models default to None."""
        state = TransportStateData(
            ID="TransportState_1",
            description="Transport state 1",
            time_model_id="tm1",
            type=StateTypeEnum.TransportState,
        )
        assert state.loading_time_model_id is None
        assert state.unloading_time_model_id is None

    def test_with_loading_and_unloading_time_models(self):
        """Test creating TransportStateData with loading/unloading time models."""
        state = TransportStateData(
            ID="TransportState_1",
            description="Transport state 1",
            time_model_id="tm1",
            type=StateTypeEnum.TransportState,
            loading_time_model_id="tm_load",
            unloading_time_model_id="tm_unload",
        )
        assert state.loading_time_model_id == "tm_load"
        assert state.unloading_time_model_id == "tm_unload"


class TestSetupStateData:
    """Tests for SetupStateData."""

    def test_creation_with_required_fields(self):
        """Test creating SetupStateData with required fields."""
        state = SetupStateData(
            ID="Setup_State_2",
            description="Setup state 2",
            time_model_id="tm1",
            type=StateTypeEnum.SetupState,
            origin_setup="P2",
            target_setup="P1",
        )
        assert state.ID == "Setup_State_2"
        assert state.time_model_id == "tm1"
        assert state.type == StateTypeEnum.SetupState
        assert state.origin_setup == "P2"
        assert state.target_setup == "P1"

    def test_hash(self, sample_production_system, sample_time_model, sample_process):
        """Test hash method."""
        state1 = SetupStateData(
            ID="Setup_State_1",
            description="Setup state 1",
            time_model_id="tm1",
            type=StateTypeEnum.SetupState,
            origin_setup="P2",
            target_setup="P1",
        )
        state2 = SetupStateData(
            ID="Setup_State_2",
            description="Setup state 2",
            time_model_id="tm1",
            type=StateTypeEnum.SetupState,
            origin_setup="P2",
            target_setup="P1",
        )
        
        hash1 = state1.hash(sample_production_system)
        hash2 = state2.hash(sample_production_system)
        
        # Same parameters should produce same hash
        assert hash1 == hash2


class TestChargingStateData:
    """Tests for ChargingStateData."""

    def test_creation_with_required_fields(self):
        """Test creating ChargingStateData with required fields."""
        state = ChargingStateData(
            ID="ChargingState_1",
            description="Charging state 1",
            time_model_id="tm1",
            type=StateTypeEnum.ChargingState,
            battery_time_model_id="tm_battery",
        )
        assert state.ID == "ChargingState_1"
        assert state.description == "Charging state 1"
        assert state.time_model_id == "tm1"
        assert state.type == StateTypeEnum.ChargingState
        assert state.battery_time_model_id == "tm_battery"

    def test_hash(self, sample_production_system, sample_time_model):
        """Test hash method."""
        # Add battery time model to production system
        battery_time_model = FunctionTimeModelData(
            ID="tm_battery",
            description="Battery time model",
            distribution_function=FunctionTimeModelEnum.Exponential,
            location=240.0,
            scale=0.0,
        )
        sample_production_system.time_model_data.append(battery_time_model)
        
        state1 = ChargingStateData(
            ID="ChargingState_1",
            description="Charging state 1",
            time_model_id="tm1",
            type=StateTypeEnum.ChargingState,
            battery_time_model_id="tm_battery",
        )
        state2 = ChargingStateData(
            ID="ChargingState_2",
            description="Charging state 2",
            time_model_id="tm1",
            type=StateTypeEnum.ChargingState,
            battery_time_model_id="tm_battery",
        )
        
        hash1 = state1.hash(sample_production_system)
        hash2 = state2.hash(sample_production_system)
        
        # Same parameters should produce same hash
        assert hash1 == hash2




"""
Analytics module for processing simulation results.

This module provides modular analytics calculations for production system simulation results.
"""

from prodsys.analytics.base import AnalyticsContext
from prodsys.analytics.data_preparation import DataPreparation
from prodsys.analytics.throughput import ThroughputAnalytics
from prodsys.analytics.scrap import ScrapAnalytics

__all__ = [
    "AnalyticsContext",
    "DataPreparation",
    "ThroughputAnalytics",
    "ScrapAnalytics",
]




#!/usr/bin/env python3
"""
Example demonstrating a manufacturing cell (SystemResource) with a robot and two machines.

This script shows:
1. How to create a SystemResource representing a manufacturing cell
2. A robot that loads/unloads products to/from the cell ports
3. Two machines that can process products (either one can be used - parallel choice)
4. ProcessModel with adjacency matrix showing parallel process execution
5. AGV transport in the global system
"""

import prodsys.express as psx


def main():
    """Main example function."""
    print("=== System Resource with Robot and Machines Example ===\n")
    
    # ========== TIME MODELS ==========
    print("1. Creating time models...")
    
    # Robot time models
    tm_robot_handling = psx.FunctionTimeModel(
        distribution_function="normal",
        location=5.0,
        scale=1.0,
        ID="tm_robot_load"
    )
    
    # Machine time models
    tm_machine1 = psx.FunctionTimeModel(
        distribution_function="normal",
        location=1.8,
        scale=0.2,
        ID="tm_machine1"
    )
    tm_machine2 = psx.FunctionTimeModel(
        distribution_function="normal",
        location=2.2,
        scale=0.7,
        ID="tm_machine2"
    )

    tm_machine3 = psx.FunctionTimeModel(
        distribution_function="normal",
        location=2.0,
        scale=0.5,
        ID="tm_machine3"
    )
    
    # Transport and system time models
    tm_agv = psx.DistanceTimeModel(
        speed=60.0,
        reaction_time=0.1,
        metric="manhattan",
        ID="tm_agv"
    )
    tm_arrival = psx.FunctionTimeModel(
        distribution_function="exponential",
        location=2.5,
        ID="tm_arrival"
    )
    
    print("   ✓ Time models created")
    
    # ========== PROCESSES ==========
    print("\n2. Creating processes...")
    
    # Robot processes
    robot_handling_process = psx.TransportProcess(
        time_model=tm_robot_handling,
        ID="robot_handling"
    )
    
    # Machine processes
    machine1_process = psx.ProductionProcess(
        time_model=tm_machine1,
        ID="machine1_process"
    )
    machine2_process = psx.ProductionProcess(
        time_model=tm_machine2,
        ID="machine2_process"
    )

    machine3_process = psx.ProductionProcess(
        time_model=tm_machine3,
        ID="machine3_process"
    )
    
    # Transport process
    agv_transport = psx.TransportProcess(
        time_model=tm_agv,
        ID="agv_transport"
    )
    
    print("   ✓ Individual processes created")
    
    # ========== PROCESS MODEL ==========
    print("\n3. Creating ProcessModel with parallel execution...")
    
    # Create ProcessModel with adjacency matrix
    # Workflow: robot_load → (machine1 OR machine2) → robot_unload
    # The adjacency matrix allows parallel choice between machine1 and machine2
    cell_process_model = psx.ProcessModel(
        adjacency_matrix={
            "machine1_process": ["machine2_process"],
            "machine2_process": []
        },
        ID="cell_process_model"
    )
    
    print("   ✓ ProcessModel created with adjacency matrix:")
    print(f"      {cell_process_model.adjacency_matrix}")

    product_process_model = psx.ProcessModel(
        can_contain_other_models=True,
        ID="product_process_model",
        adjacency_matrix={
            "machine3_process": ["cell_process_model"],
            "cell_process_model": []
        }
    )
    
    # ========== RESOURCES ==========
    print("\n4. Creating resources...")
    
    # Robot resource for loading/unloading
    robot = psx.Resource(
        # processes=[robot_handling_process], # TODO: use later capability processes hor multiple different processes
        processes=[agv_transport],
        location=[12, 10],  # Inside the cell
        capacity=1,
        ID="robot"
    )
    
    # Machine resources
    machine1 = psx.Resource(
        processes=[machine1_process],
        location=[12, 8],  # Inside the cell
        capacity=1,
        ID="machine1"
    )
    machine2 = psx.Resource(
        processes=[machine2_process],
        location=[12, 12],  # Inside the cell
        capacity=1,
        ID="machine2"
    )

    machine3 = psx.Resource(
        processes=[machine3_process],
        location=[12, 16],  # Inside the cell
        capacity=1,
        ID="machine3"
    )
    
    # AGV for global transport
    agv = psx.Resource(
        processes=[agv_transport],
        location=[0, 10],
        capacity=1,
        ID="agv"
    )
    
    print("   ✓ Robot, machines, and AGV created")
    
    # ========== SYSTEM RESOURCE (CELL) ==========
    print("\n5. Creating SystemResource (manufacturing cell)...")
    
    # Create the manufacturing cell as a SystemResource
    # The cell contains the robot and two machines
    manufacturing_cell = psx.SystemResource(
        processes=[cell_process_model],
        location=[10, 10],
        subresource_ids=["robot", "machine1", "machine2"],
        capacity=5,  # One product can be in the system at a time
        ID="manufacturing_cell"
    )
    
    print("   ✓ Manufacturing cell (SystemResource) created")
    print(f"      Subresources: {manufacturing_cell.subresource_ids}")
    
    # ========== PRODUCT ==========
    print("\n6. Creating product...")
    
    product = psx.Product(
        process=product_process_model,
        transport_process=agv_transport,
        ID="product"
    )
    
    print("   ✓ Product created with ProcessModel")
    
    # ========== SOURCES AND SINKS ==========
    print("\n7. Creating sources and sinks...")
    
    source = psx.Source(product, tm_arrival, [0, 10], ID="source")
    
    sink = psx.Sink(product, [20, 10], ID="sink")
    
    print("   ✓ Source and sink created")
    
    # ========== PRODUCTION SYSTEM ==========
    print("\n8. Creating ProductionSystem...")
    
    system = psx.ProductionSystem(
        resources=[robot, machine1, machine2, machine3, manufacturing_cell, agv],
        sources=[source],
        sinks=[sink],
        ID="production_system"
    )
    
    print("   ✓ ProductionSystem created")
    
    # ========== VALIDATION AND SIMULATION ==========
    print("\n9. Validating and running simulation...")
    
    system.validate()
    print("   ✓ System validated successfully")
    system_data = system.to_model()
    system_data.write("examples/modelling_and_simulation/simulation_example_data/system_resource_with_robot_example.json")
    
    # Run simulation
    system.run(time_range=1000)
    print("   ✓ Simulation completed")
    
    # Print results
    print("\n10. Simulation Results:")
    system.runner.print_results()
    system.runner.save_results_as_csv()

if __name__ == "__main__":
    main()




"""
`process` module contains the `prodsys.express` classes to represent the processes that can
be performed on products by resources.

The following processes are possible:
- `ProductionProcess`: A process that can be performed on a product by a production resource.
- `CapabilityProcess`: A process that can be performed on a product by a resource, based on the capability of the resource.
- `TransportProcess`: A process that can be performed on a product by a transport resource.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Optional, Union, Literal
from uuid import uuid1

from abc import ABC

from pydantic import Field, field_validator, model_validator
from pydantic.dataclasses import dataclass
from pydantic_core import ArgsKwargs

from prodsys.express import core, time_model
from prodsys.models import processes_data


@dataclass
class Process(ABC):
    """
    Abstract base class to represents a process.

    Args:
        time_model (time_model.TIME_MODEL_UNION): Time model of the process.

    Attributes:
        type (processes_data.ProcessTypeEnum): Type of the process.
    """

    time_model: time_model.TIME_MODEL_UNION


@dataclass
class DefaultProcess(Process):
    """
    Abstract base class to represents a process, with no additional attributes than type and ID.

    Args:
        time_model (time_model.TIME_MODEL_UNION): Time model of the process.
        ID (str): ID of the process.

    Attributes:
        type (processes_data.ProcessTypeEnum): Type of the process.
    """

    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))
    type: processes_data.ProcessTypeEnum = Field(init=False)
    dependencies: Optional[List[Dependency]] = Field(default_factory=list)


@dataclass
class ProductionProcess(DefaultProcess, core.ExpressObject):
    """
    Class that represents a production process.

    Args:
        time_model (time_model.TIME_MODEL_UNION): Time model of the process.
        failure_rate (Optional[float]): Failure rate of the process.
        ID (str): ID of the process.

    Attributes:
        type (processes_data.ProcessTypeEnum): Type of the process. Equals to processes_data.ProcessTypeEnum.ProductionProcesses.

    Examples:
        Production process with a function time model:
        ``` py
        import prodsys.express as psx
        welding_time_model = psx.time_model_data.FunctionTimeModel(
            distribution_function="normal",
            location=20.0,
            scale=5.0,
        )
        psx.ProductionProcess(
            time_model=welding_time_model,
        )
        ```
    """

    failure_rate: Optional[float] = None
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))
    type: processes_data.ProcessTypeEnum = Field(
        init=False, default=processes_data.ProcessTypeEnum.ProductionProcesses
    )

    def to_model(self) -> processes_data.ProductionProcessData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            processes_data.ProductionProcessData: Data object of the express object.
        """
        return processes_data.ProductionProcessData(
            time_model_id=self.time_model.ID,
            ID=self.ID,
            description="",
            type=self.type,
            failure_rate=self.failure_rate,
            dependency_ids=[dependency.ID for dependency in self.dependencies],
        )


@dataclass
class CapabilityProcess(Process, core.ExpressObject):
    """
    Class that represents a capability process. For capability processes, matching of
    required processes of product and provided processes by resources is done based on
    the capability instead of the porcess itself.

    Args:
        time_model (time_model.TIME_MODEL_UNION): Time model of the process.
        failure_rate (Optional[float]): Failure rate of the process.
        capability (str): Capability of the process.
        ID (str): ID of the process.

    Attributes:
        type (processes_data.ProcessTypeEnum): Type of the process. Equals to processes_data.ProcessTypeEnum.CapabilityProcesses.

    Examples:
        Capability process with a function time model:
        ``` py
        import prodsys.express as psx
        welding_time_model = psx.FunctionTimeModel(
            distribution_function="normal",
            location=20.0,
            scale=5.0,
        )
        psx.CapabilityProcess(
            time_model=welding_time_model,
            capability="welding"
        )
        ```
    """

    capability: str
    failure_rate: Optional[float] = None
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))
    type: processes_data.ProcessTypeEnum = Field(
        init=False, default=processes_data.ProcessTypeEnum.CapabilityProcesses
    )
    dependencies: Optional[List[Dependency]] = Field(default_factory=list)

    def to_model(self) -> processes_data.CapabilityProcessData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            processes_data.CapabilityProcessData: Data object of the express object.
        """
        return processes_data.CapabilityProcessData(
            time_model_id=self.time_model.ID,
            capability=self.capability,
            ID=self.ID,
            description="",
            type=self.type,
            failure_rate=self.failure_rate,
            dependency_ids=[dependency.ID for dependency in self.dependencies],
        )


@dataclass
class ReworkProcess(Process, core.ExpressObject):
    """
    Class that represents a rework process. Rework processes are required to rework a product.

    Args:
        time_model (time_model.TIME_MODEL_UNION): Time model of the process.
        reworked_process (Optional[Union[ProductionProcess, CapabilityProcess]]): Process that is reworked.
        blocking (Optional[bool]): If the rework process is blocking.
        ID (str): ID of the process.
    """

    reworked_processes: list[Union[ProductionProcess, CapabilityProcess]]
    blocking: Optional[bool] = False
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))
    type: processes_data.ProcessTypeEnum = Field(
        init=False, default=processes_data.ProcessTypeEnum.ReworkProcesses
    )
    dependencies: Optional[List[Dependency]] = Field(default_factory=list)

    def to_model(self):
        return processes_data.ReworkProcessData(
            ID=self.ID,
            description="",
            type=self.type,
            time_model_id=self.time_model.ID,
            reworked_process_ids=[
                reworked_process.ID for reworked_process in self.reworked_processes
            ],
            blocking=self.blocking,
            dependency_ids=[dependency.ID for dependency in self.dependencies],
        )


@dataclass
class TransportProcess(DefaultProcess, core.ExpressObject):
    """
    Class that represents a transport process. Transport processes are required to transport product from one location to another.

    Args:
        time_model (time_model.TIME_MODEL_UNION): Time model of the process.
        ID (str): ID of the process.
        loading_time_model (Optional[time_model.TIME_MODEL_UNION]): Time model of the loading process.
        unloading_time_model (Optional[time_model.TIME_MODEL_UNION]): Time model of the unloading process.
        type (processes_data.ProcessTypeEnum): Type of the process.

    Attributes:
        type (processes_data.ProcessTypeEnum): Type of the process. Equals to processes_data.ProcessTypeEnum.TransportProcesses.

    Examples:
        Transport process with a manhattan distance time model:
        ```py
        import prodsys.express as psx
        manhattan_time_model = psx.ManhattenDistanceTimeModel(
            speed=30.0,
            reaction_time=0.15,
        )
        psx.TransportProcess(
            time_model=manhattan_time_model
        )
        ```
    """

    loading_time_model: Optional[time_model.TIME_MODEL_UNION] = None
    unloading_time_model: Optional[time_model.TIME_MODEL_UNION] = None
    type: processes_data.ProcessTypeEnum = Field(
        init=False, default=processes_data.ProcessTypeEnum.TransportProcesses
    )

    def to_model(self) -> processes_data.TransportProcessData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            processes_data.TransportProcessData: Data object of the express object.
        """
        return processes_data.TransportProcessData(
            time_model_id=self.time_model.ID,
            ID=self.ID,
            description="",
            type=self.type,
            loading_time_model_id=(
                self.loading_time_model.ID if self.loading_time_model else None
            ),
            unloading_time_model_id=(
                self.unloading_time_model.ID if self.unloading_time_model else None
            ),
            dependency_ids=[dependency.ID for dependency in self.dependencies],
        )


@dataclass
class LinkTransportProcess(TransportProcess):
    """
    Represents a link transport process. They include a list or dict of links.

    Attributes:
        type (processes_data.ProcessTypeEnum): The type of the process.
        ID (Optional[str]): The ID of the process.
        links (Union[List[List[Union[resources.Resource, resources.NodeData, source.Source, sink.Sink]]],
                      Dict[Union[resources.Resource, resources.NodeData, source.Source, sink.Sink],
                           List[Union[resources.Resource, resources.NodeData, source.Source, sink.Sink]]]]):
            The links associated with the process. Each link is a list of different objects, which can be a
            Resource, NodeData, Source, or Sink. If the links attribute is a list, it represents a list of
            links, where each link is a list of these objects. If the links attribute is a dictionary, it represents a
            mapping from a key (which can be a Resource, NodeData, Source, or Sink) to a list of these objects.
        capability (Optional[str]): The capability of the process.
    """

    type: processes_data.ProcessTypeEnum = Field(
        init=False, default=processes_data.ProcessTypeEnum.LinkTransportProcesses
    )
    links: Union[
        List[List[Union[resources.Resource, sink.Sink, source.Source, node.Node]]],
        Dict[
            Union[resources.Resource, sink.Sink, source.Source, node.Node],
            List[Union[resources.Resource, sink.Sink, source.Source, node.Node]],
        ],
    ] = Field(default_factory=list)
    capability: Optional[str] = Field(default_factory=str)

    def add_link(
        self, link: List[Union[resources.Resource, sink.Sink, source.Source, node.Node]]
    ) -> None:
        """
        Adds a link to the LinkTransportProcess object.

        Args:
            link (Union[Resource, Node, Source, Sink]): The link to add.
            link_list (List[Union[Resource, Node, Source, Sink]]): The list of links to add the link to.
        """
        if isinstance(self.links, list):
            self.links.append(link)
        else:
            if not link in self.links:
                self.links[link] = []
            self.links[link[0]] = link[1]

    def set_links(
        self,
        links: List[
            List[Union[resources.Resource, node.Node, source.Source, sink.Sink]]
        ],
    ) -> None:
        """
        Sets the links of the LinkTransportProcess object.

        Args:
            links (List[List[Union[Resource, Node, Source, Sink]]]): The links to set.
        """
        self.links = links

    def to_model(self) -> processes_data.LinkTransportProcessData:
        """
        Converts the LinkTransportProcess object to its corresponding data model.

        Returns:
            processes_data.LinkTransportProcessData: The converted data model object.
        """
        if isinstance(self.links, list):
            return_links = [[link.ID for link in link_list] for link_list in self.links]
        else:
            return_links = []
            for start, targets in self.links.items():
                for target in targets:
                    return_links.append([start.ID, target.ID])

        return processes_data.LinkTransportProcessData(
            time_model_id=self.time_model.ID,
            ID=self.ID,
            description="",
            type=self.type,
            links=return_links,
            capability=self.capability,
            loading_time_model_id=(
                self.loading_time_model.ID if self.loading_time_model else None
            ),
            unloading_time_model_id=(
                self.unloading_time_model.ID if self.unloading_time_model else None
            ),
            dependency_ids=[dependency.ID for dependency in self.dependencies],
        )


@dataclass
class RequiredCapabilityProcess(core.ExpressObject):
    """
    Represents a required capability process. A capability which can be matched with the capability of another process with a capability.

    Attributes:
        ID (Optional[str]): The ID of the process.
        type (processes_data.ProcessTypeEnum): The type of the process.
        capability (Optional[str]): The capability required by the process.
    """

    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))
    type: processes_data.ProcessTypeEnum = Field(
        init=False, default=processes_data.ProcessTypeEnum.RequiredCapabilityProcesses
    )
    capability: Optional[str] = Field(default_factory=str)

    def to_model(self) -> processes_data.RequiredCapabilityProcessData:
        """
        Converts the RequiredCapabilityProcess object to its corresponding data model.

        Returns:
            processes_data.RequiredCapabilityProcessData: The data model representation of the process.
        """
        return processes_data.RequiredCapabilityProcessData(
            ID=self.ID,
            description="",
            type=self.type,
            capability=self.capability,
        )


@dataclass
class ProcessModel(core.ExpressObject):
    """
    Class that represents a process model. A process model can model sequences of processes as a directed acyclic graph (DAG).

    Args:
        process_ids (List[str]): List of process IDs that are part of this process model.
        adjacency_matrix (Dict[str, List[str]]): Adjacency matrix representing the DAG structure.
        can_contain_other_models (bool): Whether this process model can contain other process models.
        ID (str): ID of the process model.
        processes (List[PROCESS_UNION]): List of process instances that are part of this process model.

    Examples:
        Process model with a simple sequence:
        ```py
        import prodsys.express as psx
        psx.ProcessModel(
            process_ids=["P1", "P2", "P3"],
            adjacency_matrix={"P1": ["P2"], "P2": ["P3"], "P3": []}
        )
        ```
    """
    adjacency_matrix: Dict[str, List[str]]
    processes: List["PROCESS_UNION"] = Field(default_factory=list)
    can_contain_other_models: bool = False
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))
    type: processes_data.ProcessTypeEnum = Field(
        init=False, default=processes_data.ProcessTypeEnum.ProcessModels
    )
    dependencies: Optional[List[Dependency]] = Field(default_factory=list)

    def add_process(self, process_instance: "PROCESS_UNION") -> None:
        """
        Adds a single process to the list of contained processes.
        """
        if process_instance not in self.processes:
            self.processes.append(process_instance)

    def add_processes(self, process_instances: List["PROCESS_UNION"]) -> None:
        """
        Adds multiple processes to the list of contained processes.
        """
        for process_instance in process_instances:
            self.add_process(process_instance)

    def get_contained_processes(self) -> List["PROCESS_UNION"]:
        """
        Returns all processes that are directly contained within this process model.
        """
        return list(self.processes)

    def get_all_nested_processes(self) -> List["PROCESS_UNION"]:
        """
        Returns all processes contained in this process model, including processes contained
        in child process models.
        """
        nested_processes: List["PROCESS_UNION"] = []
        for process_instance in self.processes:
            nested_processes.append(process_instance)
            if isinstance(process_instance, ProcessModel):
                nested_processes.extend(process_instance.get_all_nested_processes())
        return nested_processes

    def to_model(self) -> processes_data.ProcessModelData:
        """
        Converts the ProcessModel object to its corresponding data model.

        Returns:
            processes_data.ProcessModelData: The converted data model object.
        """
        return processes_data.ProcessModelData(
            ID=self.ID,
            description="",
            type="ProcessModels",
            adjacency_matrix=self.adjacency_matrix,
            can_contain_other_models=self.can_contain_other_models,
            dependency_ids=[dependency.ID for dependency in self.dependencies],
        )


@dataclass
class SequentialProcess(ProcessModel):
    """
    Class that represents a sequential process. A sequential process is a container for a sequence of processes.

    Args:
        process_ids (List[str]): List of process IDs that are executed sequentially.
        ID (str): ID of the sequential process.

    Examples:
        Sequential process with three processes:
        ```py
        import prodsys.express as psx
        psx.SequentialProcess(
            process_ids=["P1", "P2", "P3"]
        )
        ```
    """

    adjacency_matrix: Optional[Dict[str, List[str]]] = None
    process_ids: Optional[List[str]] = None
    type: processes_data.ProcessTypeEnum = Field(
        init=False, default=processes_data.ProcessTypeEnum.SequentialProcesses
    )

    def __post_init__(self):
        """Generate adjacency matrix for sequential execution after initialization."""
        if self.processes and self.process_ids is None:
            self.process_ids = [process_instance.ID for process_instance in self.processes]
        elif self.processes and self.process_ids is not None:
            process_ids_from_processes = [
                process_instance.ID for process_instance in self.processes
            ]
            if process_ids_from_processes != self.process_ids:
                raise ValueError(
                    "Provided process_ids do not match the IDs of the supplied processes."
                )

        if self.adjacency_matrix is None:
            if not self.process_ids:
                raise ValueError(
                    "SequentialProcess requires either process_ids or processes to be provided."
                )
            adjacency_matrix = {}
            for i, process_id in enumerate(self.process_ids):
                if i < len(self.process_ids) - 1:
                    adjacency_matrix[process_id] = [self.process_ids[i + 1]]
                else:
                    adjacency_matrix[process_id] = []
            self.adjacency_matrix = adjacency_matrix

    @model_validator(mode='before')
    @classmethod
    def validate_adjacency_matrix(cls, v):
        if v is None:
            return {}
        if ArgsKwargs and isinstance(v, ArgsKwargs):
            v = dict(v.kwargs or {})  # ignore positional args for this model
        processes = v.get("processes")
        if processes and "process_ids" not in v:
            v["process_ids"] = [process_instance.ID for process_instance in processes]
        # check that its purely sequential
        count_nodes_without_successors = 0
        start_node_count = {}
        end_node_count = {}
        if isinstance(v, dict) and "process_ids" in v and "adjacency_matrix" not in v:
            # adjacency_matrix = {process_id: [process_id + 1] for counter,process_id in enumerate(v["process_ids"]) if counter < len(v["process_ids"]) - 1}
            adjacency_matrix = {}
            for i, process_id in enumerate(v["process_ids"]):
                if i < len(v["process_ids"]) - 1:
                    adjacency_matrix[process_id] = [v["process_ids"][i + 1]]
                else:
                    adjacency_matrix[process_id] = []
            v["adjacency_matrix"] = adjacency_matrix
        else:
            adjacency_matrix = v["adjacency_matrix"]
        for process_id, successors in adjacency_matrix.items():
            if len(successors) > 1:
                raise ValueError(f"Process {process_id} has multiple successors: {successors}")
            
            if len(successors) == 0:
                count_nodes_without_successors += 1
                continue
            if process_id == successors[0]:
                raise ValueError(f"Process {process_id} has itself as successor")
            if process_id not in start_node_count:
                start_node_count[process_id] = 0
            start_node_count[process_id] += 1
            if successors[0] not in end_node_count:
                end_node_count[successors[0]] = 0
            end_node_count[successors[0]] += 1
        if count_nodes_without_successors > 1:
            raise ValueError(f"There are {count_nodes_without_successors} nodes without successors")
        if any(x > 1 for x in start_node_count.values()):
            raise ValueError(f"There are {sum(start_node_count.values())} nodes with multiple successors, probably due to a directed cycle")
        if any(x > 1 for x in end_node_count.values()):
            raise ValueError(f"There are {sum(end_node_count.values())} nodes with multiple predecessors")
        return v


@dataclass
class LoadingProcess(Process, core.ExpressObject):
    """
    Class that represents a loading process. Loading processes can be chained in sequential processes or be mandatory dependencies.

    Args:
        time_model (time_model.TIME_MODEL_UNION): Time model of the loading process.
        dependency_type (Literal["before", "after", "parallel"]): Type of dependency relationship.
        can_be_chained (bool): Whether this loading process can be chained with others.
        ID (str): ID of the loading process.

    Examples:
        Loading process with "before" dependency:
        ```py
        import prodsys.express as psx
        time_model = psx.FunctionTimeModel("normal", 2.0, 0.5)
        psx.LoadingProcess(
            time_model=time_model,
            dependency_type="before",
            can_be_chained=True
        )
        ```
    """

    dependency_type: Literal["before", "after", "parallel"]
    can_be_chained: bool = True
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))
    type: processes_data.ProcessTypeEnum = Field(
        init=False, default=processes_data.ProcessTypeEnum.LoadingProcesses
    )
    dependencies: Optional[List[Dependency]] = Field(default_factory=list)

    def to_model(self) -> processes_data.LoadingProcessData:
        """
        Converts the LoadingProcess object to its corresponding data model.

        Returns:
            processes_data.LoadingProcessData: The converted data model object.
        """
        return processes_data.LoadingProcessData(
            time_model_id=self.time_model.ID,
            ID=self.ID,
            description="",
            type=self.type,
            dependency_type=self.dependency_type,
            can_be_chained=self.can_be_chained,
            dependency_ids=[dependency.ID for dependency in self.dependencies],
        )


PROCESS_UNION = Union[
    ProductionProcess,
    CapabilityProcess,
    TransportProcess,
    RequiredCapabilityProcess,
    LinkTransportProcess,
    ReworkProcess,
    ProcessModel,
    SequentialProcess,
    LoadingProcess,
]

# Import at the end to avoid circular imports
from prodsys.express import resources, sink, source, node
from prodsys.express.dependency import Dependency
